"""
The CUDAGenerator is responsible for the complete code generation process
in ANNarchy to support CUDA devices. Generate the header for a Population
object to run either on a Nvidia GPU using Nvidia SDK > 5.0 and CC > 2.0

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""
import re
import ANNarchy

from copy import deepcopy

from ANNarchy.core import Global
from ANNarchy.core.Population import Population
from ANNarchy.core.PopulationView import PopulationView

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_precision
from ANNarchy.intern import Messages

from ANNarchy.generator.Utils import generate_equation_code, tabify, check_and_apply_pow_fix, determine_idx_type_for_projection
from ANNarchy.generator.Population.PopulationGenerator import PopulationGenerator
from ANNarchy.generator.Projection.ProjectionGenerator import ProjectionGenerator, get_bounds
from ANNarchy.generator.Projection.CUDA import *

class CUDAGenerator(ProjectionGenerator):
    """
    As stated in module description, inherits from ProjectionGenerator
    and implements abstract functions.
    """
    def __init__(self, cuda_version, profile_generator, net_id):
        # The super here calls all the base classes, so first
        # ProjectionGenerator and afterwards CUDAConnectivity
        super(CUDAGenerator, self).__init__(profile_generator, net_id)
        self._cuda_version = cuda_version

    def header_struct(self, proj, annarchy_dir):
        """
        Generate the codes for the pop[id].hpp file. This file contains
        the c-style structure with all data members and equation codes (in
        case of openMP).
        """
        # Initial state
        self._templates = deepcopy(BaseTemplates.cuda_templates)
        self._template_ids = {}

        # Select the C++ connectivity template
        sparse_matrix_include, sparse_matrix_format, sparse_matrix_args, single_matrix = self._select_sparse_matrix_format(proj, suppress_printouts=False)

        # configure Connectivity base class
        self._configure_template_ids(proj)

        # Initialize launch configuration
        init_launch_config, update_launch_config = self._generate_launch_config(proj)

        # Generate declarations and accessors for the variables
        decl, accessor = self._declaration_accessors(proj, single_matrix)

        # concurrent streams
        decl['cuda_stream'] = BaseTemplates.cuda_stream

        # Initiliaze the projection
        init_weights, init_delays, init_parameters_variables = self._init_parameters_variables(proj, single_matrix)

        variables_body, variables_invoke, variables_header, variables_call = self._update_synapse(proj)

        # Update the random distributions
        init_rng = self._init_random_distributions(proj)

        # Post event
        post_event_body, post_event_invoke, post_event_header, post_event_call = self._post_event(proj)

        # Compute sum is the trickiest part
        psp_device_kernel, psp_invoke_kernel, psp_kernel_decl, psp_host_call =\
            self._computesum_rate(proj) if proj.synapse_type.type == 'rate' else self._computesum_spiking(proj)

        # Detect event-driven variables
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True

        # Detect delays to eventually generate the code
        has_delay = proj.max_delay > 1

        # Connectivity template
        if 'declare_connectivity_matrix' not in proj._specific_template.keys():
            connector_call = self._connectivity_init(proj, sparse_matrix_format, sparse_matrix_args) % {
                'sparse_format': sparse_matrix_format,
                'init_weights': init_weights,
                'init_delays': init_delays,
                'rng_idx': "[0]" if single_matrix else "",
                'add_args': "",
                'num_threads': "",
                'float_prec': ConfigManager().get('precision', self._net_id),
                'idx_type': determine_idx_type_for_projection(proj)[0]
            }
            declare_connectivity_matrix = ""
            access_connectivity_matrix = ""
        else:
            sparse_matrix_format = "SpecificConnectivity"
            sparse_matrix_args = ""
            sparse_matrix_include = "#include \"Specific.hpp\"\n"
            connector_call = ""
            declare_connectivity_matrix = proj._specific_template['declare_connectivity_matrix']
            access_connectivity_matrix = proj._specific_template['access_connectivity_matrix']

        # Memory transfers
        host_device_transfer, device_host_transfer = self._memory_transfers(proj)

        # Memory management
        size_in_bytes = self._size_in_bytes(proj)
        clear_container = self._clear_container(proj)

        # Local functions
        host_local_func, device_local_func = self._local_functions(proj)
        decl['parameters_variables'] += host_local_func

        # Profiling
        if self._prof_gen:
            include_profile = """#include "Profiling.hpp"\n"""
            declare_profile, init_profile = self._prof_gen.generate_init_projection(proj)
        else:
            include_profile = ""
            init_profile = ""
            declare_profile = ""

        # Additional info (overwritten)
        include_additional = ""
        struct_additional = ""
        init_additional = ""
        access_additional = ""

        if 'include_additional' in proj._specific_template.keys():
            include_additional = proj._specific_template['include_additional']
        if 'struct_additional' in proj._specific_template.keys():
            struct_additional = proj._specific_template['struct_additional']
        if 'init_additional' in proj._specific_template.keys():
            init_additional = proj._specific_template['init_additional']
        if 'access_additional' in proj._specific_template.keys():
            access_additional = proj._specific_template['access_additional']

        final_code = self._templates['projection_header'] % {
            # version tag
            'annarchy_version': ANNarchy.__release__,
            # fill code templates
            'id_pre': proj.pre.id,
            'id_post': proj.post.id,
            'id_proj': proj.id,
            'name_pre': proj.pre.name,
            'name_post': proj.post.name,
            'target': proj.target,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'sparse_matrix_include': sparse_matrix_include,
            'sparse_format': sparse_matrix_format,
            'sparse_format_args': sparse_matrix_args,
            'include_additional': include_additional,
            'include_profile': include_profile,
            'struct_additional': struct_additional,
            'declare_connectivity_matrix': declare_connectivity_matrix,
            'access_connectivity_matrix': access_connectivity_matrix,
            'declare_delay': decl['declare_delay'] if has_delay else "",
            'declare_event_driven': decl['event_driven'] if has_event_driven else "",
            'declare_rng': decl['rng'],
            'declare_parameters_variables': decl['parameters_variables'],
            'declare_cuda_stream': decl['cuda_stream'],
            'declare_additional': decl['additional'],
            'declare_profile': declare_profile,
            'connector_call': connector_call,
            'init_weights': init_weights,
            'init_event_driven': "",
            'init_rng': init_rng,
            'init_launch_config': init_launch_config,
            'update_launch_config': update_launch_config,
            'init_parameters_variables': init_parameters_variables,
            'init_additional': init_additional,
            'init_profile': init_profile,
            'access_parameters_variables': accessor,
            'access_additional': access_additional,
            'host_to_device': host_device_transfer,
            'device_to_host': device_host_transfer,
            'size_in_bytes': size_in_bytes,
            'clear_container': clear_container
        }

        # Store the file in generate ( will be compared with files contained
        # in build/ later on )
        with open(annarchy_dir+'/generate/net'+str(self._net_id)+'/proj'+str(proj.id)+'.hpp', 'w') as ofile:
            ofile.write(final_code)

        # Build dictionary for inclusions in ANNarchy.cu
        proj_desc = {
            'include': """#include "proj%(id)s.hpp"\n""" % {'id': proj.id},
            'extern': """extern ProjStruct%(id)s* proj%(id)s;\n"""% {'id': proj.id},
            'instance': """ProjStruct%(id)s* proj%(id)s;\n"""% {'id': proj.id},
            'init': """    proj%(id)s->init_projection();\n""" % {'id' : proj.id}
        }

        # synaptic transmission: continuous / pre-event
        proj_desc['psp_device_kernel'] = psp_device_kernel
        proj_desc['psp_invoke_kernel'] = psp_invoke_kernel
        proj_desc['psp_kernel_decl'] = psp_kernel_decl
        proj_desc['psp_host_call'] = psp_host_call

        proj_desc['update_synapse_header'] = variables_header
        proj_desc['update_synapse_body'] = variables_body
        proj_desc['update_synapse_invoke'] = variables_invoke
        proj_desc['update_synapse_call'] = variables_call

        proj_desc['postevent_header'] = post_event_header
        proj_desc['postevent_body'] = post_event_body
        proj_desc['postevent_invoke'] = post_event_invoke
        proj_desc['postevent_call'] = post_event_call

        proj_desc['custom_func'] = device_local_func

        proj_desc['host_to_device'] = tabify("proj%(id)s->host_to_device();" % {'id':proj.id}, 1)+"\n"
        proj_desc['device_to_host'] = tabify("proj%(id)s->device_to_host();" % {'id':proj.id}, 1)+"\n"

        return proj_desc

    def _configure_template_ids(self, proj):
        """
        Assign the correct template dictionary based on projection
        storage format.
        """
        # HD (13th Dec. 2022): for now the data-type is ignored in almost
        #                      all templates related to spiking models
        idx_type, _, size_type, _ = determine_idx_type_for_projection(proj)

        # Some common ids
        self._template_ids.update({
            'id_proj' : proj.id,
            'target': proj.target,
            'id_post': proj.post.id,
            'id_pre': proj.pre.id,
            'idx_type': idx_type,
            'size_type': size_type,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'pre_prefix': 'pre_',
            'post_prefix': 'post_',
            'host_pre_prefix': 'pop'+ str(proj.pre.id) + '->',
            'host_post_prefix': 'pop'+ str(proj.post.id) + '->'
        })

        # Indices to access data depend on the storage format/storage order
        # Please note, that the indices stored in the template files are related
        # to rate-coded models/codes.
        if proj._storage_format == "csr":
            if proj._storage_order == "post_to_pre":
                self._templates.update(CSR_CUDA.conn_templates)
                self._template_ids.update(CSR_CUDA.conn_ids)
            else:
                self._templates.update(CSR_T_CUDA.conn_templates)
                self._template_ids.update(CSR_T_CUDA.conn_ids)

        elif proj._storage_format == "csr_scalar":
            self._templates.update(CSR_SCALAR_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(CSR_SCALAR_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "csr_vector":
            self._templates.update(CSR_VECTOR_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(CSR_VECTOR_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "bsr":
            self._templates.update(BSR_CUDA.conn_templates)
            self._template_ids.update(BSR_CUDA.conn_ids)

        elif proj._storage_format == "coo":
            self._templates.update(COO_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(COO_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "sell":
            self._templates.update(SELL_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(SELL_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "ellr":
            self._templates.update(ELLR_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(ELLR_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "ell":
            self._templates.update(ELL_CUDA.conn_templates)
            if proj._storage_order == "post_to_pre":
                self._template_ids.update(ELL_CUDA.conn_ids)
            else:
                raise NotImplementedError

        elif proj._storage_format == "hyb":
            self._templates.update(HYB_CUDA.conn_templates)
            # Indices must be set locally for each part

        elif proj._storage_format == "dense":
            if proj._storage_order == "post_to_pre":
                self._templates.update(Dense_CUDA.conn_templates)
                self._template_ids.update(Dense_CUDA.conn_ids)
            else:
                self._templates.update(Dense_T_CUDA.conn_templates)
                self._template_ids.update(Dense_T_CUDA.conn_ids)

        else:
            raise Messages.InvalidConfiguration("   The storage_format="+str(proj._storage_format)+" is not available on CUDA devices")

    def _generate_launch_config(self, proj):
        """
        TODO: multiple targets???
        """
        if isinstance(proj, SpecificProjection):
            # HD (24th Feb. 2023): user-defined codes has their own kernel-call statements,
            #                      therefore this adjustment is not needed
            return "", ""

        init_code = self._templates['launch_config']['init'] % {
            'id_proj': proj.id
        }
        update_code = self._templates['launch_config']['update'] % {
            'id_proj': proj.id
        }
        return init_code, update_code

    def _computesum_rate(self, proj):
        """
        Returns the code templates needed to compute post-synaptic sum kernels:

        device_kernel:     device kernel
        invoke_kernel:     kernel invocation
        kernel_decl:   kernel export
        host_call:              call invoke function 
        """
        # Specific projection
        if 'psp_header' in proj._specific_template.keys() or \
            'psp_body' in proj._specific_template.keys() or \
            'psp_call' in proj._specific_template.keys():

            try:
                device_kernel_header = proj._specific_template['psp_header']
                kernel_invoke = proj._specific_template['psp_invoke']
                device_kernel = proj._specific_template['psp_body']
                host_call = proj._specific_template['psp_call']
            except KeyError:
                Messages._error('At least one of the following fields is missing for psp: header, invoke, body or call')

            if self._prof_gen:
                host_call = self._prof_gen.annotate_computesum_rate(proj, host_call)

            return device_kernel, kernel_invoke, device_kernel_header, host_call

        # Dictionary of keywords to transform the parsed equations
        ids = deepcopy(self._template_ids)

        # Some adjustments to spare single used local variables
        if proj._storage_format == "ellr":
            ids['post_index'] = "[rank_post[i]]"
            ids['pre_index'] = "[rank_pre[j*post_size+i]]"
        elif proj._storage_format == "bsr":
            ids['pre_prefix'] = "loc_pre_"
            ids['pre_index'] = "[col]"

        # Dependencies
        dependencies = list(set(proj.synapse_type.description['dependencies']['pre']))

        #
        # Retrieve the PSP
        add_args_header = ""
        add_args_call = ""
        add_args_kernel = ""
        if not 'psp' in  proj.synapse_type.description.keys(): # default
            psp = """%(preprefix)sr%(pre_index)s * w%(local_index)s;"""
            add_args_header += "const %(float_prec)s* __restrict__ pre_r, const %(float_prec)s* __restrict__ w" % {'float_prec': ConfigManager().get('precision', self._net_id)}
            add_args_call = "pop%(id_pre)s->gpu_r, proj%(id_proj)s->gpu_w " % {'id_proj': proj.id, 'id_pre': proj.pre.id}
            add_args_kernel = "pre_r, w"

        else: # custom psp
            psp = (proj.synapse_type.description['psp']['cpp'])

            # update dependencies
            for dep in proj.synapse_type.description['psp']['dependencies']:
                _, attr = self._get_attr_and_type(proj, dep)
                attr_ids = {
                    'id_proj': proj.id,
                    'type': attr['ctype'],
                    'name': attr['name']
                }
                add_args_header += ", const %(type)s* __restrict__ %(name)s" % attr_ids
                add_args_call += ", proj%(id_proj)s->gpu_%(name)s" % attr_ids
                add_args_kernel += ", %(name)s" % attr_ids

            for dep in list(set(proj.synapse_type.description['dependencies']['pre'])):
                _, attr = PopulationGenerator._get_attr_and_type(proj.pre, dep)
                attr_ids = {
                    'id_pre': proj.pre.id,
                    'type': attr['ctype'],
                    'name': attr['name'],
                    'host_pre_prefix': self._template_ids['host_pre_prefix']
                }

                add_args_header += ", const %(type)s* __restrict__ pre_%(name)s" % attr_ids
                add_args_call += ", %(host_pre_prefix)sgpu_%(name)s" % attr_ids
                add_args_kernel += ", pre_%(name)s" % attr_ids

        # Special case where w is a single value
        if proj._has_single_weight():
            psp = re.sub(
                r'([^\w]+)w%\(local_index\)s',
                r'\1w',
                ' ' + psp
            )

        # Allow the use of global variables in psp (issue60)
        for var in dependencies:
            if var in proj.pre.neuron_type.description['global']:
                psp = psp.replace("%(pre_prefix)s"+var+"%(pre_index)s", "%(pre_prefix)s"+var+"%(global_index)s")

        # connectivity, yet only CSR
        conn_header = ""
        conn_call = ""

        #
        # finish the kernel etc.
        operation = proj.synapse_type.operation

        if proj._storage_format != "hyb":
            # connectivity
            conn_header = self._templates['conn_header'] % ids
            conn_call = self._templates['conn_call'] % ids
            conn_kernel = self._templates['conn_kernel']

            body_dict = deepcopy(ids)
            body_dict.update({
                # device function
                'conn_args': conn_header,
                'target_arg': "sum_"+proj.target,
                'add_args': add_args_header,
                'psp': psp  % ids,
                'thread_init': self._templates['rate_psp']['thread_init'][ConfigManager().get('precision', self._net_id)][operation],
                # call function
                'conn_args_call': conn_kernel,
                'target_arg_call': ", sum_%(target)s" % {'id_post': proj.post.id, 'target': proj.target},
                'add_args_call': add_args_kernel,
            })
            device_kernel = self._templates['rate_psp']['device_kernel'][operation] % body_dict

            if 'invoke_kernel' in self._templates['rate_psp'].keys():
                invoke_kernel = self._templates['rate_psp']['invoke_kernel'] % body_dict
                kernel_decl = self._templates['rate_psp']['kernel_decl'] % body_dict
            else:
                invoke_kernel = ""
                kernel_decl = self._templates['rate_psp']['kernel_decl'] % {
                    'float_prec': ConfigManager().get('precision', self._net_id),
                    'id': proj.id,
                    'conn_args': conn_header,
                    'target_arg': "sum_"+proj.target,
                    'add_args': add_args_header
                }

            host_call_dict = deepcopy(ids)
            host_call_dict.update({
                'conn_args': conn_call,
                'target_arg': ", pop%(id_post)s->gpu__sum_%(target)s" % host_call_dict,
                'add_args': add_args_call
            })
            host_call = self._templates['rate_psp']['host_call'] % host_call_dict

        else:
            # Should be equal to ProjectionGenerator._configure_template_ids()
            idx_type, _, size_type, _ = determine_idx_type_for_projection(proj)
            id_dict = {
                'id_proj': proj.id,
                'id_pre': proj.pre.id,
                'id_post': proj.post.id,
                'idx_type': idx_type,
                'size_type': size_type
            }

            #
            # ELLPACK - partition
            conn_header = ELL_CUDA.conn_templates['conn_header'] % id_dict
            conn_call = ELL_CUDA.conn_templates['conn_call'] % id_dict
            ell_ids = {
                'idx_type': idx_type,
                'local_index': "[j*post_size+i]",
                'semiglobal_index': '[i]',
                'global_index': '[0]',
                'pre_index': '[rank_pre[j*post_size+i]]',
                'post_index': '[rank_post[i]]',
                'pre_prefix': 'pre_',
                'post_prefix': 'post_'
            }
            body_code = ELL_CUDA.conn_templates['rate_psp']['body'][operation] % {
                'idx_type': idx_type,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'id_proj': proj.id,
                'conn_args': conn_header,
                'target_arg': "sum_"+proj.target,
                'add_args': add_args_header,
                'psp': psp  % ell_ids,
                'thread_init': ELLR_CUDA.conn_templates['rate_psp']['thread_init'][ConfigManager().get('precision', self._net_id)][operation],
                'post_index': ell_ids['post_index']
            }
            header_code = ELL_CUDA.conn_templates['rate_psp']['header'] % {
                'idx_type': idx_type,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'id': proj.id,
                'conn_args': conn_header,
                'target_arg': "sum_"+proj.target,
                'add_args': add_args_header
            }

            #
            # Coordinate - partition
            conn_header = COO_CUDA.conn_templates['conn_header'] % id_dict
            conn_call = COO_CUDA.conn_templates['conn_call'] % id_dict
            coo_ids = {
                'local_index': "[j]",
                'semiglobal_index': '[i]',
                'global_index': '[0]',
                'pre_index': '[column_indices[j]]',
                'post_index': '[row_indices[j]]',
                'pre_prefix': 'pre_',
                'post_prefix': 'post_',
            }
            body_code += COO_CUDA.conn_templates['rate_psp']['body'][operation] % {
                'float_prec': ConfigManager().get('precision', self._net_id),
                'idx_type': idx_type,
                'size_type': size_type,
                'id_proj': proj.id,
                'conn_args': conn_header,
                'target_arg': "sum_"+proj.target,
                'add_args': add_args_header,
                'psp': psp  % coo_ids,
                'thread_init': COO_CUDA.conn_templates['rate_psp']['thread_init'][ConfigManager().get('precision', self._net_id)][operation],
                'post_index': coo_ids['post_index']
            }
            header_code += COO_CUDA.conn_templates['rate_psp']['header'] % {
                'float_prec': ConfigManager().get('precision', self._net_id),
                'id': proj.id,
                'conn_args': conn_header,
                'target_arg': "sum_"+proj.target,
                'add_args': add_args_header
            }

            # update dependencies
            add_args_call_coo = add_args_call
            add_args_call_ell = add_args_call
            for dep in proj.synapse_type.description['psp']['dependencies']:
                add_args_call_coo = add_args_call_coo.replace("gpu_"+dep+",", "gpu_"+dep+"->coo,")
                add_args_call_ell = add_args_call_ell.replace("gpu_"+dep+",", "gpu_"+dep+"->ell,")

            call_code = HYB_CUDA.conn_templates['rate_psp']['call'] % {
                'id_proj': proj.id,
                'id_pre': proj.pre.id,
                'id_post': proj.post.id,
                'conn_args': conn_call,
                'target': proj.target,
                'target_arg': ", pop%(id_post)s->gpu__sum_%(target)s" % {'id_post': proj.post.id, 'target': proj.target},
                'add_args_coo': add_args_call_coo,
                'add_args_ell': add_args_call_ell,
                'float_prec': ConfigManager().get('precision', self._net_id)
            }

        # Take delays into account if any
        if proj.max_delay > 1:
            # Delayed variables
            if isinstance(proj.pre, PopulationView):
                delayed_variables = proj.pre.population.delayed_variables
            else:
                delayed_variables = proj.pre.delayed_variables

            id_pre = str(proj.pre.id)
            for var in sorted(list(set(delayed_variables))):
                host_call = host_call.replace("pop"+id_pre+"->gpu_"+var, "pop"+id_pre+"->gpu_delayed_"+var+"[proj"+str(proj.id)+"->delay-1]")

        # Profiling
        if self._prof_gen:
            host_call = self._prof_gen.annotate_computesum_rate(proj, host_call)

        return device_kernel, invoke_kernel, kernel_decl, host_call

    def _computesum_spiking(self, proj):
        """
        Generate code for the spike propagation. As ANNarchy supports a set of
        different data structures, this method split in up into several sub
        functions.

        In contrast to _computesum_rate() the spike propagation kernel need
        to implement the signal transmission (event- as well as continous)
        and also the equations filled in the 'pre-spike' field of synapse
        desctiption.
        """
        if 'psp_header' in proj._specific_template.keys() or \
            'psp_body' in proj._specific_template.keys() or \
            'psp_call' in proj._specific_template.keys():

            try:
                device_kernel_header = proj._specific_template['psp_header']
                kernel_invoke = proj._specific_template['psp_invoke']
                device_kernel = proj._specific_template['psp_body']
                host_call = proj._specific_template['psp_call']
            except KeyError:
                Messages._error('At least one of the following fields is missing for psp: header, invoke, body or call')

            if self._prof_gen:
                host_call = self._prof_gen.annotate_computesum_rate(proj, host_call)

            return device_kernel, kernel_invoke, device_kernel_header, host_call

        # some variables needed for the final templates
        psp_code = ""
        kernel_args_header = ""
        kernel_args_invoke = ""
        kernel_args_call = ""

        pre_spike_code = ""
        kernel_deps = []

        if proj.max_delay > 1 and proj.uniform_delay == -1:
            Messages._error("Non-uniform delays are not supported yet on GPUs.")

        # Basic tags, dependent on storage format are assuming a feedforward
        # transmission.
        ids = deepcopy(self._template_ids)

        # The spike transmission is triggered from pre-synaptic side
        # and the indices need to be changed.
        # HD (13th Dec. 2022): I implement this for now as in SingleThread-/OpenMP-
        #                      generator. On the long-term, we should have a second
        #                      set of conn ids? (TODO - HD/JV)
        if proj._storage_format == "csr":
            if proj._storage_order == "post_to_pre":
                ids.update({
                    'local_index': "[syn_idx]",
                    'semiglobal_index': '[post_rank]',
                    'global_index': '[0]',
                    'float_prec': ConfigManager().get('precision', self._net_id),
                    'pre_index': '[row_idx[syn_idx]]',
                    'post_index': '[post_rank]',
                })
            else:
                ids.update({
                    'local_index': "[syn_idx]",
                    'pre_index': '[col_idx[syn_idx]]'
                })

        elif proj._storage_format == "dense":
            # HD (27th Feb. 2023): the dense matrix has no explicit indices, as the position is
            #                      computed inside the kernel.
            if proj._storage_order == "post_to_pre":
                pass
            else:
                pass

        else:
            # just a reminder to check indices for new formats
            raise Messages.CodeGeneratorException("\tno indices defined for spiking psp_template and storage_format = "+ proj._storage_format)

        #
        # All statements in the 'pre_spike' field of synapse description
        #
        for var in proj.synapse_type.description['pre_spike']:
            if var['name'] == "g_target":   # synaptic transmission
                # compute psp
                psp_code += "%(float_prec)s tmp = %(psp)s\n" % {
                    'psp': var['cpp'].split('=')[1] % ids,
                    'float_prec': ConfigManager().get('precision', self._net_id)
                }
                # Operation (g_target is replaced by sum in 'cpp')
                operation = re.search(r'sum (.*?)=', var['cpp']).group(1).strip() + "="
                if operation == "+=":
                    ids.update({'atomicOp': "atomicAdd"})
                elif operation == "-=":
                    ids.update({'atomicOp': "atomicSub"})
                elif operation == "=":
                    ids.update({'atomicOp': "atomicExch"})
                else:
                    Messages._error("The operator '"+operation+"' is not supported in psp-statements on CUDA devices yet.")

                # apply to all targets
                target_list = proj.target if isinstance(proj.target, list) else [proj.target]
                for target in sorted(list(set(target_list))):
                    # multiple targets
                    ids['target'] = target

                    # Check for special cases
                    if ids['atomicOp'] == "atomicExch" and _check_precision("double", self._net_id):
                        # HD (12th April 2023): atomicExch is not defined for double, so we need to use the long long type-cast version
                        psp_code += "atomicExch((unsigned long long int*)&g_%(target)s%(post_index)s, __double_as_longlong(tmp));\n" % ids
                    else:
                        psp_code += "%(atomicOp)s(&g_%(target)s%(post_index)s, tmp);\n" % ids

                    # Boundary code is optional
                    bound_code = get_bounds(var)
                    if len(bound_code) != 0:
                        bound_code = bound_code.replace("g_target%(local_index)s", "g_"+target+"%(post_index)s")
                        psp_code += bound_code % ids

            else:
                condition = ""
                # Check conditions to update the variable
                if var['name'] == 'w': # Surround it by the learning flag
                    condition = "plasticity"

                # Flags avoids pre-spike evaluation when post fires at the same time
                if 'unless_post' in var['flags']:
                    simultaneous = "pop%(id_pre)s_last_spike[_pr] != pop%(id_post)s_last_spike[%(semiglobal_index)s]" % ids
                    if condition == "":
                        condition = simultaneous
                    else:
                        condition += "&&(" + simultaneous + ")"

                eq_dict = {
                    'eq': var['eq'],
                    'cpp': var['cpp'],
                    'bounds': get_bounds(var),
                    'condition': condition,
                }

                # Generate the code
                if condition != "":
                    pre_spike_code += """
// unless_post can prevent evaluation of presynaptic variables
if(%(condition)s){
    // %(eq)s
    %(cpp)s
    %(bounds)s
}
""" % eq_dict
                else: # Normal synaptic variable
                    pre_spike_code += """
// %(eq)s
%(cpp)s
%(bounds)s""" % eq_dict

            # Update the dependencies
            kernel_deps.append(var['name']) # right side
            for dep in var['dependencies']: # left side
                kernel_deps.append(dep)

        #
        # Event-driven integration of synaptic variables
        #
        has_exact = False
        event_driven_code = ''
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_exact = True
                event_dict = {
                    'eq': var['eq'],
                    'exact': var['cpp'].replace('(t)', '(t-1)')
                }
                event_driven_code += """
    // %(eq)s
    %(exact)s
""" % event_dict
                # add the dependencies to kernel dependencies
                for dep in var['dependencies']:
                    kernel_deps.append(dep)

        # Does an event-driven variable occur?
        if has_exact:
            event_driven_code += """
    // Update the last event for the synapse
    _last_event%(local_index)s = t;
"""

            # event-driven requires access to last event variable
            kernel_args_header += ", long* _last_event"
            kernel_args_invoke += ", _last_event"
            kernel_args_call += ", proj%(id_proj)s->_gpu_last_event"  % ids

        # Add pre- and post-synaptic population dependencies
        pre_post_deps = list(set(proj.synapse_type.description['dependencies']['pre'] + proj.synapse_type.description['dependencies']['post']))
        pre_post_args = self._gen_kernel_args(proj, pre_post_deps, pre_post_deps)
        kernel_args_header += pre_post_args[0]
        kernel_args_invoke += pre_post_args[1]
        kernel_args_call += pre_post_args[2]

        # sort out doublings
        kernel_deps = sorted(list(set(kernel_deps)))

        # Add synaptic variables to kernel arguments
        for dep in kernel_deps:
            if dep == "w" or dep == "g_target":
                # already contained
                continue

            attr_type, attr_dict = self._get_attr_and_type(proj, dep)
            attr_ids = {
                'id_proj': proj.id,
                'name': attr_dict['name'],
                'type': attr_dict['ctype']
            }

            if attr_type == 'par' and attr_dict['locality'] == "global":
                kernel_args_header += ", const %(type)s %(name)s" % attr_ids
                kernel_args_invoke += ", %(name)s" % attr_ids
                kernel_args_call += ", proj%(id_proj)s->%(name)s" % attr_ids

                # replace any occurences of this parameter
                if event_driven_code.strip() != '':
                    event_driven_code = event_driven_code.replace(attr_dict['name']+'%(global_index)s', attr_dict['name'])
                if pre_spike_code.strip() != '':
                    pre_spike_code = pre_spike_code.replace(attr_dict['name']+'%(global_index)s', attr_dict['name'])
            else:
                kernel_args_header += ", %(type)s* %(name)s" % attr_ids
                kernel_args_invoke += ", %(name)s" % attr_ids
                kernel_args_call += ", proj%(id_proj)s->gpu_%(name)s" % attr_ids

        # PSP targets
        targets_call = ""
        targets_invoke = ""
        targets_header = ""
        target_list = proj.target if isinstance(proj.target, list) else [proj.target]
        for target in target_list:
            targets_call += ", pop%(id_post)s->gpu_g_"+target
            targets_invoke += ", g_"+target
            targets_header += (", %(float_prec)s* g_"+target) % {'float_prec': ConfigManager().get('precision', self._net_id)}

        # Construct code for event-driven transmission
        #
        if len(pre_spike_code) == 0 and len(psp_code) == 0:
            # no event-driven component detected
            device_kernel = ""
            invoke_kernel = ""
            kernel_decl = ""
            host_call= ""

        else:
            # select the correct template
            template = self._templates['spike_transmission']['event_driven']

            # Connectivity description, we need to read-out the view
            # which represents the pre-synaptic entries which means
            # columns in post-to-pre and rows for pre-to-post orientation.
            if proj._storage_format == "csr":
                if proj._storage_order == "post_to_pre":
                    conn_args_header = "%(size_type)s* col_ptr, %(idx_type)s* row_idx, %(idx_type)s* inv_idx, %(float_prec)s *w" % ids
                    conn_args_invoke = "col_ptr, row_idx, inv_idx, w"
                    conn_call = "proj%(id_proj)s->gpu_col_ptr, proj%(id_proj)s->gpu_row_idx, proj%(id_proj)s->gpu_inv_idx, proj%(id_proj)s->gpu_w" % ids
                else:
                    conn_args_header = "%(size_type)s* row_ptr, %(idx_type)s* col_idx, %(float_prec)s *w" % ids
                    conn_args_invoke = "row_ptr, col_idx, w"
                    conn_call = "proj%(id_proj)s->gpu_row_ptr, proj%(id_proj)s->gpu_col_idx, proj%(id_proj)s->gpu_w" % ids
            elif proj._storage_format == "dense":
                conn_args_header = "const %(idx_type)s row_size, const %(idx_type)s column_size, %(float_prec)s *w" % ids
                conn_args_invoke = "row_size, column_size, w"
                conn_call = "proj%(id_proj)s->num_rows(), proj%(id_proj)s->num_columns(), proj%(id_proj)s->gpu_w" % ids
            else:
                raise NotImplementedError

            # Population sizes
            pre_size = proj.pre.size if isinstance(proj.pre, Population) else proj.pre.population.size
            post_size = proj.post.size if isinstance(proj.post, Population) else proj.post.population.size

            # Delays
            if proj.max_delay > 1:
                if proj.uniform_delay == -1: # Non-uniform delays
                    raise NotImplementedError
                else:
                    pre_spike_events = "pop%(id_pre)s->gpu_delayed_spiked[proj%(id_proj)s->delay-1]" % {'id_pre': proj.pre.id, 'id_proj': proj.id}
                    pre_spike_count = "pop%(id_pre)s->host_delayed_num_events[proj%(id_proj)s->delay-1]" % {'id_pre': proj.pre.id, 'id_proj': proj.id}
            else:
                pre_spike_events = "pop%(id_pre)s->gpu_spiked" % {'id_pre': proj.pre.id}
                pre_spike_count = "pop%(id_pre)s->spike_count" % {'id_pre': proj.pre.id}

            idx_type, _, size_type, _ = determine_idx_type_for_projection(proj)

            # Finalize event-driven part
            device_kernel = template['device_kernel'] % {
                'id_proj': proj.id,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'conn_args_header': conn_args_header + targets_header,
                'kernel_args_header': kernel_args_header,
                'event_driven': tabify(event_driven_code % ids, 2),
                'psp': tabify(psp_code, 3),
                'pre_event': tabify(pre_spike_code % ids, 3),
                'pre_size': pre_size,
                'post_size': post_size,
                'target': target_list[0],  # only for dense!
                'idx_type': idx_type,
                'size_type': size_type
            }
            invoke_kernel = template['invoke_kernel'] % {
                'id_proj': proj.id,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'conn_args_header': conn_args_header + targets_header,
                'conn_args_invoke': conn_args_invoke + targets_invoke,
                'kernel_args_header': kernel_args_header,
                'kernel_args_invoke': kernel_args_invoke,
                'pre_spike_events': pre_spike_events.replace("pop"+str(proj.pre.id)+"->gpu_", ""),
                'pre_spike_count': pre_spike_count.replace("pop"+str(proj.pre.id)+".", ""),
            }
            kernel_decl = template['kernel_decl'] % {
                'id_proj': proj.id,
                'float_prec': ConfigManager().get('precision', self._net_id),
                'conn_args_header': conn_args_header + targets_header,
                'kernel_args_header': kernel_args_header
            }
            host_call = template['host_call'] % {
                'id_proj': proj.id,
                'id_pre': proj.pre.id,
                'id_post': proj.post.id,
                'target': target_list[0],
                'pre_spike_events': pre_spike_events,
                'pre_spike_count': pre_spike_count,
                'kernel_args': kernel_args_call  % {'id_post': proj.post.id, 'target': target},
                'conn_args': conn_call + targets_call % {'id_post': proj.post.id}
            }

        # If the synaptic transmission is not event-based,
        # we need to add a rate-coded-like kernel.
        if 'psp' in  proj.synapse_type.description.keys():
            # transfrom psp equation
            psp_code = proj.synapse_type.description['psp']['cpp']

            if proj._storage_format == "csr":
                if proj._storage_order == "post_to_pre":
                    ids.update({
                        'pre_index': '[col_idx[syn_idx]]',
                        'post_index': '[post_rank]',
                    })
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

            # update dependencies
            for dep in proj.synapse_type.description['psp']['dependencies']:
                if dep == "w":
                    continue

                _, attr = self._get_attr_and_type(proj, dep)
                attr_ids = {
                    'id_proj': proj.id,
                    'type': attr['ctype'],
                    'name': attr['name']
                }
                kernel_args_header += ", %(type)s* %(name)s" % attr_ids
                kernel_args_invoke += ", %(name)s" % attr_ids
                kernel_args_call += ", proj%(id_proj)s->gpu_%(name)s" % attr_ids

            psp_code = proj.synapse_type.description['psp']['cpp'] % ids

            # select the correct template
            template = self._templates['spike_transmission']['continuous']

            host_call += template['host_call'] % {
                'id_proj': proj.id,
                'id_pre': proj.pre.id,
                'id_post': proj.post.id,
                'target_arg': targets_call % {'id_post': proj.post.id},
                'target': proj.target,
                'kernel_args': kernel_args_call,
                'float_prec': ConfigManager().get('precision', self._net_id)
            }
            device_kernel += template['device_kernel'] % {
                'id_proj': proj.id,
                'target_arg': target_list[0],
                'kernel_args': kernel_args_header,
                'psp': psp_code,
                'pre_code': tabify(pre_spike_code % ids, 3),
                'float_prec': ConfigManager().get('precision', self._net_id)
            }
            invoke_kernel += template['invoke_kernel'] % {
                'id_proj': proj.id,
                'target_arg': proj.target,
                'kernel_args': kernel_args_header,
                'kernel_args_invoke': kernel_args_invoke,
                'target_arg': targets_header,
                'target_arg_invoke': targets_invoke,
                'float_prec': ConfigManager().get('precision', self._net_id)
            }
            kernel_decl += template['kernel_decl'] % {
                'id_proj': proj.id,
                'kernel_args': kernel_args_header,
                'target_arg': targets_header,
                'float_prec': ConfigManager().get('precision', self._net_id)
            }

        # Annotate code
        if self._prof_gen:
            host_call = self._prof_gen.annotate_computesum_spiking(proj, host_call)

        return device_kernel, invoke_kernel, kernel_decl, host_call


    def _declaration_accessors(self, proj, single_matrix):
        """
        Extend basic declaration statements by CUDA streams.
        """
        declaration, accessor = ProjectionGenerator._declaration_accessors(self, proj, single_matrix)

        declaration['cuda_stream'] = BaseTemplates.cuda_stream
        return declaration, accessor

    @staticmethod
    def _select_deps(proj, locality):
        """
        Dependencies of synaptic equations consist of several components:

        * access to pre- or post-population
        * variables / parameters of the projection
        * pre- or post-spike event

        Return:

        * syn_deps      list of all dependencies
        * neur_deps     list of dependencies part of neurons
        """
        syn_deps = []

        # Access to pre- or postsynaptic neurons
        neur_deps = list(set(proj.synapse_type.description['dependencies']['pre'] +
                             proj.synapse_type.description['dependencies']['post']))
        for dep in neur_deps:
            syn_deps.append(dep)

        # Variables
        for var in proj.synapse_type.description['variables']:
            if var['eq'] == '':
                continue # nothing to do here

            if var['locality'] == locality:
                syn_deps.append(var['name'])
                for dep in var['dependencies']:
                    syn_deps.append(dep)

        # Random distributions
        for rd in proj.synapse_type.description['random_distributions']:
            for dep in rd['dependencies']:
                syn_deps += dep

        syn_deps = list(set(syn_deps))
        return syn_deps, neur_deps

    @staticmethod
    def _gen_kernel_args(proj, pop_deps, deps):
        """
        The header and function definitions as well as the call statement need
        to be extended with the additional variables.

        Parameters:

        * proj: currently processed projection object
        * pop_deps: list of of variable names which are either from pre- or post-synaptic populations
        * deps: list of all attribute names which are dependencies

        Returns (3 strings): kernel_args_decl, kernel_args_invoke, kernel_args_call
        """
        kernel_args_decl = ""
        kernel_args_invoke = ""
        kernel_args_call = ""

        for dep in deps:
            # The variable dep is part of pre-/post population
            if dep in pop_deps:

                if dep in proj.synapse_type.description['dependencies']['pre']:
                    attr_type, attr_dict = PopulationGenerator._get_attr_and_type(proj.pre, dep)
                    ids = {
                        'type': attr_dict['ctype'],
                        'name': attr_dict['name'],
                        'id': proj.pre.id
                    }
                    kernel_args_decl += ", %(type)s* pre_%(name)s" % ids
                    kernel_args_invoke += ", pre_%(name)s" % ids
                    kernel_args_call += ", pop%(id)s->gpu_%(name)s" % ids

                if dep in proj.synapse_type.description['dependencies']['post']:
                    attr_type, attr_dict = PopulationGenerator._get_attr_and_type(proj.post, dep)
                    ids = {
                        'type': attr_dict['ctype'],
                        'name': attr_dict['name'],
                        'id': proj.post.id
                    }
                    kernel_args_decl += ", %(type)s* post_%(name)s" % ids
                    kernel_args_invoke += ", post_%(name)s" % ids
                    kernel_args_call += ", pop%(id)s->gpu_%(name)s" % ids

            # The variable dep is part of the projection
            else:
                attr_type, attr_dict = ProjectionGenerator._get_attr_and_type(proj, dep)

                if attr_type == "par":
                    ids = {
                        'id_proj': proj.id,
                        'type': attr_dict['ctype'],
                        'name': attr_dict['name']
                    }

                    if dep in proj.synapse_type.description['global']:
                        kernel_args_decl += ", const %(type)s %(name)s" % ids
                        kernel_args_invoke += ", %(name)s" % ids
                        kernel_args_call += ", proj%(id_proj)s->%(name)s" % ids
                    else:
                        kernel_args_decl += ", %(type)s* %(name)s" % ids
                        kernel_args_invoke += ", %(name)s" % ids
                        kernel_args_call += ", proj%(id_proj)s->gpu_%(name)s" % ids

                elif attr_type == "var":
                    ids = {
                        'id_proj': proj.id,
                        'type': attr_dict['ctype'],
                        'name': attr_dict['name']
                    }
                    kernel_args_decl += ", %(type)s* %(name)s" % ids
                    kernel_args_invoke += ", %(name)s" % ids
                    kernel_args_call += ", proj%(id_proj)s->gpu_%(name)s" % ids

                elif attr_type == "rand":
                    ids = {
                        'id_proj': proj.id,
                        'type': 'curandState',
                        'name': attr_dict['name']
                    }
                    kernel_args_decl += ", %(type)s* state_%(name)s" % ids
                    kernel_args_invoke += ", state_%(name)s" % ids
                    kernel_args_call += ", proj%(id_proj)s->gpu_%(name)s" % ids
                else:
                    raise ValueError("attr_type for variable " + dep +" is invalid")

        #
        # global operations related to pre- and post-synaptic operations
        for glop in proj.synapse_type.description['pre_global_operations']:
            attr = PopulationGenerator._get_attr(proj.pre, glop['variable'])
            ids = {
                'id': proj.pre.id,
                'name': glop['variable'],
                'type': attr['ctype'],
                'func': glop['function']
            }
            kernel_args_decl += ", %(type)s pre__%(func)s_%(name)s" % ids
            kernel_args_invoke += ", pre__%(func)s_%(name)s" % ids
            kernel_args_call += ", pop%(id)s->_%(func)s_%(name)s" % ids

        for glop in proj.synapse_type.description['post_global_operations']:
            attr = PopulationGenerator._get_attr(proj.post, glop['variable'])
            ids = {
                'id': proj.post.id,
                'name': glop['variable'],
                'type': attr['ctype'],
                'func': glop['function']
            }
            kernel_args_decl += ", %(type)s post__%(func)s_%(name)s" % ids
            kernel_args_invoke += ", post__%(func)s_%(name)s" % ids
            kernel_args_call += ", pop%(id)s->_%(func)s_%(name)s" % ids

        #
        # event-driven spike synapses require the access to last_spike member
        # of pre- and post-synaptic populations.
        if proj.synapse_type.type == "spike":
            kernel_args_decl = ", long int* pre_last_spike, long int* post_last_spike" + kernel_args_decl
            kernel_args_invoke = ", pre_last_spike, post_last_spike" + kernel_args_invoke
            kernel_args_call = ", pop%(id_pre)s->gpu_last_spike, pop%(id_post)s->gpu_last_spike" % {'id_pre': proj.pre.id, 'id_post': proj.post.id} + kernel_args_call

        return kernel_args_decl, kernel_args_invoke, kernel_args_call

    def _header_structural_plasticity(self, proj):
        Messages._error("Structural Plasticity is not supported on GPUs yet.")

    def _local_functions(self, proj):
        """
        Definition of user-defined local functions attached to
        a neuron. These functions will take place in the
        ANNarchyDevice.cu file.

        As the local functions can be occur repeatadly in the same file,
        there are modified with pop[id]_ to unique them.

        Return:

            * host_define, device_define
        """
        # Local functions
        if len(proj.synapse_type.description['functions']) == 0:
            return "", ""

        host_code = ""
        device_code = ""
        for func in proj.synapse_type.description['functions']:
            cpp_func = func['cpp'] + '\n'

            host_code += cpp_func
            # TODO: improve code
            if _check_precision('float', self._net_id):
                device_code += cpp_func.replace('float' + func['name'], '__device__ float proj%(id)s_%(func)s' % {'id': proj.id, 'func': func['name']})
            else:
                device_code += cpp_func.replace('double '+ func['name'], '__device__ double proj%(id)s_%(func)s' % {'id': proj.id, 'func':func['name']})

        return host_code, check_and_apply_pow_fix(device_code, self._cuda_version)

    def _replace_local_funcs(self, proj, glob_eqs, semiglobal_eqs, loc_eqs):
        """
        As the local functions can be occur repeatadly in the same file,
        there are modified with proj[id]_ to unique them. Now we need
        to adjust the call accordingly.
        """
        for func in proj.synapse_type.description['functions']:
            search_term = r"%(name)s\([^\(]*\)" % {'name': func['name']}

            func_occur = re.findall(search_term, glob_eqs)
            for term in func_occur:
                glob_eqs = loc_eqs.replace(term, term.replace(func['name'], 'proj'+str(proj.id)+'_'+func['name']))

            func_occur = re.findall(search_term, semiglobal_eqs)
            for term in func_occur:
                semiglobal_eqs = loc_eqs.replace(term, term.replace(func['name'], 'proj'+str(proj.id)+'_'+func['name']))

            func_occur = re.findall(search_term, loc_eqs)
            for term in func_occur:
                loc_eqs = loc_eqs.replace(term, term.replace(func['name'], 'proj'+str(proj.id)+'_'+func['name']))

        return glob_eqs, semiglobal_eqs, loc_eqs

    def _replace_random(self, loc_eqs, loc_idx, glob_eqs, random_distributions):
        """
        This method replace the variables rand_%(id)s in the parsed equations
        by the corresponding curand... term.
        """
        # double precision methods have a postfix
        prec_extension = "" if _check_precision('float', self._net_id) else "_double"

        loc_pre = ""
        semi_pre = ""
        glob_pre = ""

        for dist in random_distributions:
            print(dist)
            print(loc_eqs)
            if dist['dist'] == "Uniform":
                dist_ids = {
                    'postfix': prec_extension,
                    'rd': dist['name'],
                    'min': dist['args'].split(',')[0],
                    'max': dist['args'].split(',')[1],
                    'local_index': loc_idx
                }

                if dist["locality"] == "local":
                    term = """( curand_uniform%(postfix)s( &state_%(rd)s%(local_index)s ) * (%(max)s - %(min)s) + %(min)s )""" % dist_ids
                    loc_pre += "%(prec)s %(name)s = %(term)s;" % {'prec': ConfigManager().get('precision', self._net_id), 'name': dist['name'], 'term': term}

                    # suppress local index
                    loc_eqs = loc_eqs.replace(dist['name']+loc_idx, dist['name'])
                else:
                    # HD (17th May 2021): this path can not be reached as the parser rejects equations like:
                    # dw/dt = -w * Uniform(0,.1) : init=1, midpoint
                    raise NotImplementedError

            elif dist['dist'] == "Normal":
                dist_ids = {
                    'postfix': prec_extension, 'rd': dist['name'],
                    'mean': dist['args'].split(",")[0],
                    'sigma': dist['args'].split(",")[1]
                }

                if dist["locality"] == "local":
                    term = """( curand_normal%(postfix)s( &state_%(rd)s[j] ) * %(sigma)s + %(mean)s )""" % dist_ids
                    loc_pre += "%(prec)s %(name)s = %(term)s;" % {'prec': ConfigManager().get('precision', self._net_id), 'name': dist['name'], 'term': term}

                    # suppress local index
                    loc_eqs = loc_eqs.replace(dist['name']+"[j]", dist['name'])
                else:
                    # HD (17th May 2021): this path can not be reached as the parser rejects equations like:
                    # dw/dt = -w * Uniform(0,.1) : init=1, midpoint
                    raise NotImplementedError

            elif dist['dist'] == "LogNormal":
                dist_ids = {
                    'postfix': prec_extension, 'rd': dist['name'],
                    'mean': dist['args'].split(',')[0],
                    'std_dev': dist['args'].split(',')[1]
                }

                if dist["locality"] == "local":
                    term = """( curand_log_normal%(postfix)s( &state_%(rd)s[j], %(mean)s, %(std_dev)s) )""" % dist_ids
                    loc_pre += "%(prec)s %(name)s = %(term)s;" % {'prec': ConfigManager().get('precision', self._net_id), 'name': dist['name'], 'term': term}

                    # suppress local index
                    loc_eqs = loc_eqs.replace(dist['name']+"[j]", dist['name'])
                else:
                    # HD (17th May 2021): this path can not be reached as the parser rejects equations like:
                    # dw/dt = -w * Uniform(0,.1) : init=1, midpoint
                    raise NotImplementedError

            else:
                Messages._error("Unsupported random distribution on GPUs: " + dist['dist'])

        # check which equation blocks we need to extend
        if len(loc_pre) > 0:
            loc_eqs = tabify(loc_pre, 2) + "\n" + loc_eqs
        if len(glob_pre) > 0:
            glob_eqs = tabify(glob_pre, 1) + "\n" + glob_eqs

        return loc_eqs, glob_eqs

    def _post_event(self, proj):
        """
        Post-synaptic event kernel for CUDA devices
        """
        if proj.synapse_type.type == "rate":
            return "", "", "", ""

        if proj.synapse_type.description['post_spike'] == []:
            return "", "", "", ""

        # Get basic template ids
        ids = deepcopy(self._template_ids)

        add_args_header = ""
        add_args_invoke = ""
        add_args_call = ""

        # Event-driven integration
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True

        # Generate event-driven code
        event_driven_code = ""
        event_deps = []
        if has_event_driven:
            # event-driven rely on last pre-synaptic event
            add_args_header += ", long* _last_event"
            add_args_invoke += ", _last_event"
            add_args_call += ", proj%(id_proj)s->_gpu_last_event" % {'id_proj': proj.id}

            for var in proj.synapse_type.description['variables']:
                if var['method'] == 'event-driven':
                    event_driven_code += '// ' + var['eq'] + '\n'
                    event_driven_code += var['cpp'] + '\n'

                    for deps in var['dependencies']:
                        event_deps.append(deps)
            event_driven_code += """
// Update the last event for the synapse
_last_event%(local_index)s = t;
""" % {'local_index' : '[j]'}

        # Gather the equations and the list of dependent variables
        post_code = ""
        post_deps = []
        for post_eq in proj.synapse_type.description['post_spike']:
            post_code += '// ' + post_eq['eq'] + '\n'
            if post_eq['name'] == 'w':
                post_code += "if(plasticity)\n"
            post_code += post_eq['cpp'] + '\n'
            post_code += get_bounds(post_eq) + '\n'

            # add dependencies, only right side!
            for deps in post_eq['dependencies']:
                post_deps.append(deps)

            # left side of equations is not part of dependencies
            post_deps.append(post_eq['name'])

        # Create add_args for event-driven eqs and post_event
        kernel_deps = sorted(list(set(post_deps+event_deps))) # variables can occur in several eqs
        for dep in kernel_deps:
            if dep == "w":
                continue

            attr_type, attr_dict = self._get_attr_and_type(proj, dep)
            attr_ids = {
                'id': proj.id, 'type': attr_dict['ctype'], 'name': attr_dict['name']
            }
            if attr_type == 'par' and attr_dict['locality'] == "global":
                add_args_header += ', const %(type)s %(name)s' % attr_ids
                add_args_invoke += ', %(name)s' % attr_ids
                add_args_call += ', proj%(id)s->%(name)s' % attr_ids

                if post_code.strip != '':
                    post_code = post_code.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
                if event_driven_code.strip() != '':
                    event_driven_code = event_driven_code.replace(attr_dict['name']+"%(global_index)s", attr_dict['name'])
            else:
                add_args_header += ', %(type)s* %(name)s' % attr_ids
                add_args_invoke += ', %(name)s' % attr_ids
                add_args_call += ', proj%(id)s->gpu_%(name)s' % attr_ids

        # Check for equations which consider post-synaptic neural state variables
        for post_eq in proj.synapse_type.description['post_spike']:
            for dep in post_eq['prepost_dependencies']['pre']:
                attr_type, attr_dict = PopulationGenerator._get_attr_and_type(proj.pre, dep)
                attr_ids = {
                    'id': proj.pre.id, 'type': attr_dict['ctype'], 'name': attr_dict['name']
                }
                add_args_header += ', %(type)s* pre_%(name)s' % attr_ids
                add_args_invoke += ', pre_%(name)s' % attr_ids
                add_args_call += ', pop%(id)s->gpu_%(name)s' % attr_ids

            for dep in post_eq['prepost_dependencies']['post']:
                attr_type, attr_dict = PopulationGenerator._get_attr_and_type(proj.post, dep)
                attr_ids = {
                    'id': proj.post.id, 'type': attr_dict['ctype'], 'name': attr_dict['name']
                }
                add_args_header += ', %(type)s* post_%(name)s' % attr_ids
                add_args_invoke += ', post_%(name)s' % attr_ids
                add_args_call += ', pop%(id)s->gpu_%(name)s' % attr_ids

        # select code template
        try:
            templates = self._templates['post_event']

        except KeyError:
            raise Messages._error("No CUDA code template for post_event ( format =", proj._storage_format, " and order =", proj._storage_order,")")

        # Fill code templates
        postevent_body = templates['device_kernel'] % {
            'id_proj': proj.id,
            'conn_args': self._templates['conn_header'] % ids,
            'add_args': add_args_header,
            'event_driven': tabify(event_driven_code % ids, 2),
            'post_code': tabify(post_code % ids, 2),
            'float_prec': ConfigManager().get('precision', self._net_id),
        }

        postevent_invoke = templates['invoke_kernel'] % {
            'id_proj': proj.id,
            'conn_args': self._templates['conn_header'] % ids,
            'conn_args_invoke': self._templates['conn_kernel'],
            'add_args': add_args_header,
            'add_args_invoke': add_args_invoke,
            'float_prec': ConfigManager().get('precision', self._net_id)
        }

        postevent_header = templates['kernel_decl'] % {
            'id_proj': proj.id,
            'conn_args': self._templates['conn_header']% ids,
            'add_args': add_args_header,
            'float_prec': ConfigManager().get('precision', self._net_id)
        }

        postevent_call = templates['host_call'] % {
            'id_proj': proj.id,
            'id_pre': proj.pre.id,
            'id_post': proj.post.id,
            'target': proj.target[0] if isinstance(proj.target, list) else proj.target,
            'conn_args': self._templates['conn_call'] % ids,
            'add_args': add_args_call
        }

        # Annotate code
        if self._prof_gen:
            postevent_call = self._prof_gen.annotate_post_event(proj, postevent_call)

        return postevent_body, postevent_invoke, postevent_header, postevent_call

    def _init_random_distributions(self, proj):
        # Random numbers
        code = ""
        if len(proj.synapse_type.description['random_distributions']) > 0:
            code += """
        // Random numbers"""
            for dist in proj.synapse_type.description['random_distributions']:
                rng_ids = {
                    'id': proj.id,
                    'rd_name': dist['name'],
                }
                code += self._templates['rng'][dist['locality']]['init'] % rng_ids

        return code

    def _memory_transfers(self, proj):
        """
        Generate source code for transfer variables and parameters.
        """
        if 'host_device_transfer' in  proj._specific_template.keys() and \
            'device_host_transfer' in proj._specific_template.keys():
            return proj._specific_template['host_device_transfer'], proj._specific_template['device_host_transfer']

        host_device_transfer = ""
        device_host_transfer = ""

        proc_attr = []
        for attr in proj.synapse_type.description['parameters']+proj.synapse_type.description['variables']:
            # avoid doublons
            if attr['name'] in proc_attr:
                continue

            attr_type = "parameter" if attr in proj.synapse_type.description['parameters'] else "variable"
            locality = attr['locality']
            if attr_type == "parameter" and locality == "global":
                continue

            #
            # Host -> Device
            #
            host_device_transfer += self._templates['host_to_device'][locality] % {
                'id': proj.id,
                'name': attr['name'],
                'type': attr['ctype']
            }

            #
            # Device -> Host
            #
            device_host_transfer += self._templates['device_to_host'][locality] % {
                'id': proj.id, 'name': attr['name'], 'type': attr['ctype']
            }

            proc_attr.append(attr['name'])

        return host_device_transfer, device_host_transfer

    def _process_equations(self, proj, equations, ids, locality):
        """
        Process the equation block and create equation blocks and
        corresponding kernel argument list and call statement.

        .. Note:

        This function is a helper function and should be called by
        _update_synapse() only.
        """
        # Process equations and determine dependencies
        syn_deps, neur_deps = self._select_deps(proj, locality)
        kernel_args, kernel_args_invoke, kernel_args_call = self._gen_kernel_args(proj, neur_deps, syn_deps)

        # Add pre_rank/post_rank identifier if needed
        rk_assign = ""
        if locality == "semiglobal":
            if proj._storage_format in ["csr"] :
                rk_assign += "%(idx_type)s rk_post = rank_post%(semiglobal_index)s;\n"
            elif proj._storage_format in ["ellr"]:
                rk_assign += ""
            elif proj._storage_format in ["dense"]:
                rk_assign += "%(idx_type)s rk_post = i;\n"
            else:
                raise NotImplementedError

        elif locality=="local":
            # rk_pre/rk_pre depend on the matrix format
            if proj._storage_format in ["csr"] :
                if proj._storage_order == "post_to_pre":
                    rk_assign += "%(idx_type)s rk_pre = rank_pre%(local_index)s;\n"
                else:
                    rk_assign += "%(idx_type)s rk_post = rank_post%(semiglobal_index)s;\n"
            elif proj._storage_format in ["dense"]:
                pass
            else:
                rk_assign += "%(idx_type)s rk_post = rank_post%(semiglobal_index)s;\n"
                rk_assign += "%(idx_type)s rk_pre = rank_pre%(local_index)s;\n"

        # finalize rank assignment code
        rk_assign = tabify(rk_assign, 2 if proj._storage_format == "csr" else 3)

        # Gather pre-loop declaration (dt/tau for ODEs)
        pre_loop = ""
        for var in proj.synapse_type.description['variables']:
            if var['locality'] == locality:
                if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                    pre_loop += ConfigManager().get('precision', self._net_id) + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'
            else:
                continue

        # Global parameters have no index
        for syn_dep in syn_deps:
            attr_type, attr_dict = self._get_attr_and_type(proj, syn_dep)

            if attr_type == "par" and attr_dict['locality'] == "global" :
                equations = equations.replace(attr_dict["name"]+"%(global_index)s", attr_dict["name"])

                if pre_loop.strip() != '':
                    pre_loop = pre_loop.replace(attr_dict["name"]+"%(global_index)s", attr_dict["name"])

        # Finalize equations, add pre-loop and/or rank assignment
        equations = (rk_assign + equations) % ids
        if pre_loop.strip() != '':
            pre_loop = """\n// Updating the step sizes\n""" + pre_loop % ids

        return equations, pre_loop, kernel_args, kernel_args_invoke, kernel_args_call

    def _update_synapse(self, proj):
        """
        Generate the device codes for synaptic equations. As the parallel
        evaluation of local and global equations within one kernel would require
        a __syncthread() call, we split up the implementation into two seperate
        parts.

        Return:

        * a tuple contain three strings ( device_kernel, call, header )
        """
        # Global variables
        global_eq = generate_equation_code(proj.synapse_type.description, 'global', 'proj', padding=1, wrap_w="plasticity")

        # Semiglobal variables
        semiglobal_eq = generate_equation_code(proj.synapse_type.description, 'semiglobal', 'proj', padding=2, wrap_w="plasticity")

        # Local variables
        pad = 2 if proj._storage_format == "csr" else 3
        local_eq = generate_equation_code(proj.synapse_type.description, 'local', 'proj', padding=pad, wrap_w="plasticity")

        # Something to do?
        if global_eq.strip() == '' and semiglobal_eq.strip() == '' and local_eq.strip() == '':
            return "", "", "", ""

        # Modify the default dictionary for specific formats
        ids = deepcopy(self._template_ids)
        if proj._storage_format == "ell":
            ids['pre_index'] = '[rk_pre]'
            ids['post_index'] = '[rk_post]'
        elif proj._storage_format == "csr":
            ids['pre_index'] = '[rk_pre]'
            ids['post_index'] = '[rk_post]'

        # CPP type for indices
        idx_type, _, size_type, _ = determine_idx_type_for_projection(proj)

        # some commonly needed ids
        ids.update({
            'id_proj': proj.id,
            'id_pre': proj.pre.id,
            'id_post': proj.post.id,
            'float_prec': ConfigManager().get('precision', self._net_id),
            'idx_type': idx_type,
            'size_type': size_type
        })

        # generate the code
        device_kernel = ""
        invoke_kernel = ""
        kernel_header = ""
        local_call = ""
        global_call = ""
        semiglobal_call = ""

        #
        # Fill code templates for global, semiglobal and local equations
        #
        if global_eq.strip() != '':
            global_eq, global_pre_code, kernel_args_global, kernel_args_invoke_global, kernel_args_call_global =\
                self._process_equations( proj, global_eq, ids, 'global' )

        if semiglobal_eq.strip() != '':
            semiglobal_eq, semiglobal_pre_code, kernel_args_semiglobal, kernel_args_invoke_semiglobal, kernel_args_call_semiglobal =\
                self._process_equations( proj, semiglobal_eq, ids, 'semiglobal' )

        if local_eq.strip() != '':
            local_eq, local_pre_code, kernel_args_local, kernel_args_invoke_local, kernel_args_call_local =\
                self._process_equations( proj, local_eq, ids, 'local' )

        # replace the random distributions
        local_eq, global_eq = self._replace_random(local_eq, ids['local_index'], global_eq, proj.synapse_type.description['random_distributions'])

        #
        # replace local function calls
        if len(proj.synapse_type.description['functions']) > 0:
            global_eq, semiglobal_eq, local_eq = self._replace_local_funcs(proj, global_eq, semiglobal_eq, local_eq)

        # connectivity
        conn_header = self._templates['conn_header'] % ids
        conn_invoke = self._templates['conn_kernel'] % ids
        conn_call = self._templates['conn_call'] % ids

        # we seperated the execution of global/semiglobal/local into three kernels
        # as the threads would have two different loads.
        if global_eq.strip() != '':
            body_dict = {
                'kernel_args': kernel_args_global,
                'global_eqs': global_eq,
                'pre_loop':  global_pre_code,
            }
            body_dict.update(ids)
            device_kernel += self._templates['synapse_update']['global']['device_kernel'] % body_dict

            invoke_dict = {
                'kernel_args': kernel_args_global,
                'kernel_args_call': kernel_args_invoke_global
            }
            invoke_dict.update(ids)
            invoke_kernel += self._templates['synapse_update']['global']['invoke_kernel'] % invoke_dict

            header_dict = {
                'kernel_args': kernel_args_global,
            }
            header_dict.update(ids)
            kernel_header += self._templates['synapse_update']['global']['kernel_decl'] % header_dict

            call_dict = deepcopy(ids)
            call_dict.update({
                'target': proj.target[0] if isinstance(proj.target, list) else proj.target,
                'kernel_args_call': kernel_args_call_global,
            })
            global_call = self._templates['synapse_update']['global']['host_call'] % call_dict

        if semiglobal_eq.strip() != '':
            body_dict = {
                'kernel_args': kernel_args_semiglobal,
                'semiglobal_eqs': semiglobal_eq,
                'pre_loop': semiglobal_pre_code,
            }
            body_dict.update(ids)
            device_kernel += self._templates['synapse_update']['semiglobal']['device_kernel'] % body_dict

            invoke_dict = {
                'kernel_args': kernel_args_semiglobal,
                'kernel_args_call': kernel_args_invoke_semiglobal
            }
            invoke_dict.update(ids)
            invoke_kernel += self._templates['synapse_update']['semiglobal']['invoke_kernel'] % invoke_dict

            header_dict = {
                'kernel_args': kernel_args_semiglobal,
            }
            header_dict.update(ids)
            kernel_header += self._templates['synapse_update']['semiglobal']['kernel_decl'] % header_dict

            call_dict = deepcopy(ids)
            call_dict.update({
                'target': proj.target[0] if isinstance(proj.target, list) else proj.target,
                'kernel_args_call': kernel_args_call_semiglobal,
            })
            semiglobal_call = self._templates['synapse_update']['semiglobal']['host_call'] % call_dict

        if local_eq.strip() != '':
            body_dict = {
                'conn_args': conn_header,
                'kernel_args': kernel_args_local,
                'local_eqs': local_eq,
                'pre_loop': tabify(local_pre_code,1)
            }
            body_dict.update(ids)
            device_kernel += self._templates['synapse_update']['local']['device_kernel'] % body_dict

            invoke_dict = {
                'conn_args': conn_header,
                'conn_args_call': conn_invoke,
                'kernel_args': kernel_args_local,
                'kernel_args_call': kernel_args_invoke_local
            }
            invoke_dict.update(ids)
            invoke_kernel += self._templates['synapse_update']['local']['invoke_kernel'] % invoke_dict

            header_dict = {
                'conn_args': conn_header,
                'kernel_args': kernel_args_local
            }
            header_dict.update(ids)
            kernel_header += self._templates['synapse_update']['local']['kernel_decl'] % header_dict

            call_dict = deepcopy(ids)
            call_dict.update({
                'target': proj.target[0] if isinstance(proj.target, list) else proj.target,
                'conn_args_call': conn_call,
                'kernel_args_call': kernel_args_call_local
            })
            local_call = self._templates['synapse_update']['local']['host_call'] % call_dict

        call = self._templates['synapse_update']['call'] % {
            'id_proj': proj.id,
            'post': proj.post.id,
            'pre': proj.pre.id,
            'target': proj.target[0] if isinstance(proj.target, list) else proj.target,
            'global_call': global_call,
            'semiglobal_call': semiglobal_call,
            'local_call': local_call,
            'float_prec': ConfigManager().get('precision', self._net_id)
        }

        # Profiling
        if self._prof_gen:
            call = self._prof_gen.annotate_update_synapse(proj, call)

        return device_kernel, invoke_kernel, kernel_header, call



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.core.Random import RandomDistribution
from ANNarchy.intern import Messages

class PopulationView :
    """
    Subset of a Population.

    A list of ranks can be passed, but population views are usually created through slicing:

    ```python
    pop = net.create(1000, neuron)

    E = pop[:800] # ann.PopulationView(pop, range(800))
    I = pop[800:] # ann.PopulationView(pop, range(800, 1000))
    ```

    :param population: population object
    :param ranks: list or numpy array containing the ranks of the selected neurons.
    :param geometry: a geometry for the Populationview (optional)
    """

    def __init__(self, population, ranks, geometry=None):

        self.population = population
        "Original population."

        self.ranks = np.array(ranks)
        "Array of ranks in the PopulationView."

        self.geometry = geometry
        "Geometry of the PopulationView (optional)."

        self.size = len(self.ranks)
        "Size of the PopulationView (number of neurons)."
        
        # Internal attributes        
        self.neuron_type = self.population.neuron_type
        self.id = self.population.id
        self.offsets = [np.amin(self.ranks), np.amax(self.ranks)+1]
        self.cyInstance = population.cyInstance

    def _copy(self):
        "Returns a copy of the population when creating networks. Internal use only."
        return PopulationView(population=self.population, ranks=self.ranks, geometry=self.geometry)

    ################################
    # Indexing
    ################################

    def __len__(self):
        """
        Number of neurons in the population view.
        """
        return self.size

    def rank_from_coordinates(self, coord:tuple, local:bool=False) -> int:
        """
        Returns the rank of a neuron based on coordinates.

        When local is `False` (default), the coordinates are relative to the ORIGINAL population, not the PopulationView.

        When local is `True`, the coordinates are interpreted relative to the geometry of the PopulationView if available. When you add two population views, the geometry is lost and the method will return an error.

        The rank is relative to the original population. Iterate over len(pop) otherwise.

        :param coord: coordinate tuple, can be multidimensional.
        :param local: whether the coordinates are local to the PopulationView or not.
        """
        if not local:
            rk = self.population.rank_from_coordinates(coord)
            if not rk in self.ranks:
                Messages._error("There is no neuron of coordinates", coord, "in the PopulationView.")
            return rk

        else:
            if not self.geometry:
                Messages._error("The population view does not have a geometry, cannot use local coordinates.")
            else:
                try:
                    intern_rank = np.ravel_multi_index(coord, self.geometry)
                except:
                    Messages._error("There is no neuron of coordinates", coord, "in a PopulationView of geometry", self.geometry)
                return self.ranks[intern_rank]

    def coordinates_from_rank(self, rank:int, local:bool=False) -> tuple:
        """
        Returns the coordinates of a neuron based on its rank.

        When local is `False` (default), the coordinates are relative to the ORIGINAL population, not the PopulationView.

        When local is `True`, the coordinates are interpreted relative to the geometry of the PopulationView if available. When you add two population views, the geometry is lost and the method will return an error.

        The rank is relative to the original population. Iterate over len(pop) otherwise.

        :param rank: rank of the neuron in the original population.
        :param local: whether the coordinates are local to the PopulationView or not.
        """
        if not local:
            return self.population.coordinates_from_rank(rank)
        else:
            if not self.geometry:
                Messages._error("The population view does not have a geometry, cannot use local coordinates.")
            else:
                if not rank in self.ranks:
                    Messages._error("There is no neuron of rank", rank, "in the PopulationView.")
                intern_rk = self.ranks.index(rank)
                coord = np.unravel_index(intern_rk, self.geometry)
                return coord


    ################################
    # Targets must match the population, both in read and write
    ################################
    @property
    def targets(self) -> list[str]:
        "List of targets connected to the population."
        return self.population.targets

    @targets.setter
    def targets(self, value):
        self.population.targets.append(value)

    @property
    def name(self) -> str:
        "Name of the original population."
        return self.population.name

    @property
    def max_delay(self) -> str:
        return self.population.max_delay

    @property
    def attributes(self) -> list[str]:
        "Äist of attributes of the original population."
        return self.population.attributes

    @property
    def variables(self) -> list[str]:
        "List of variables of the original population."
        return self.population.variables

    @property
    def parameters(self) -> list[str]:
        "List of constants of the original population."
        return self.population.parameters

    ################################
    ## Access to attributes
    ################################

    def __getattr__(self, name):
        " Method called when accessing an attribute."
        if name == 'population':
            return object.__getattribute__(self, name)
        elif name == 'spike':
            all_events = set(self.population.spike)
            own_ranks = set(self.ranks)
            return list(sorted(set.intersection(all_events, own_ranks)))
        elif hasattr(self.population, 'attributes'):
            if name in self.population.attributes:
                return self.get(name)
            else:
                return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        " Method called when setting an attribute."
        if name == 'population':
            object.__setattr__(self, name, value)
        elif hasattr(self, 'population'):
            if name in self.population.attributes:
                self.set({name: value})
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def get(self, name):
        """
        Returns the parameter/variable value.

        :param name: name of the parameter/variable.
        """
        if name in self.population.attributes:
            all_val = getattr(self.population, name).reshape(self.population.size)
            return all_val[self.ranks]
        else:
            Messages._error("Population does not have a parameter/variable called " + name + ".")

    def set(self, value:dict) -> None:
        """
        Updates the neurons' parameter/variable values.

        **Warning:** If you modify the value of a global parameter, this will be the case for ALL neurons of the population, not only the subset.

        :param value: dictionary of parameters/variables to be updated for the corresponding subset of neurons. It can be a single value or a list/1D array of the same size as the `PopulationView`.
        """
        def _set_single(name, rank, value):
            if not self.population.initialized:
                if not name in self.population.neuron_type.description['local']:
                    Messages._error('can not set the value of a global attribute from a PopulationView.')
                    return

                if isinstance(self.population.init[name], np.ndarray):
                    if len(self.population.geometry) == 1:
                        self.population.init[name][rank] = value
                    else: # Need to access the coordinates
                        coords = self.population.coordinates_from_rank(rank)
                        self.population.init[name][coords] = value
                else:
                    val = self.population.init[name]
                    data = val * np.ones(self.population.size)
                    data[rank] = value
                    self.population.init[name] = data.reshape(self.population.geometry)
            else:
                var = getattr(self.population.cyInstance, name)
                var[rank] = value
                setattr(self.population.cyInstance, name, var)

        for val_key in value.keys():
            if hasattr(self.population, val_key):
                # Check the value
                if isinstance(value[val_key], RandomDistribution): # Make sure it is generated only once
                        value[val_key] = np.array(value[val_key].get_values(self.size))
                if isinstance(value[val_key], np.ndarray): # np.array
                    if value[val_key].ndim >1 or len(value[val_key]) != self.size:
                        Messages._error("You can only provide an array of the same size as the PopulationView", self.size)
                        return None
                    if val_key in self.population.neuron_type.description['global']:
                        Messages._error("Global attributes can only have one value in a population.")
                        return None
                    # Assign the value
                    for idx, rk in enumerate(self.ranks):
                        _set_single(val_key, rk, value[val_key][idx])

                elif isinstance(value[val_key], list): # list
                    if len(value[val_key]) != self.size:
                        Messages._error("You can only provide a list of the same size as the PopulationView", self.size)
                        return None
                    if val_key in self.population.neuron_type.description['global']:
                        Messages._error("Global attributes can only have one value in a population.")
                        return None
                    # Assign the value
                    for idx, rk in enumerate(self.ranks):
                        _set_single(val_key, rk, value[val_key][idx])

                else: # single value
                    for rk in self.ranks:
                        _set_single(val_key, rk, value[val_key])
            else:
                Messages._error("the population has no attribute called ", val_key)
                return None

    ################################
    ## Access to weighted sums
    ################################
    def sum(self, target:str) -> np.ndarray:
        """
        Returns the array of weighted sums corresponding to the target:

        ```python
        excitatory = pop[:50].sum('exc')
        ```

        For spiking networks, this is equivalent to accessing the conductances directly::

        ```python
        excitatory = pop[:50].g_exc
        ```

        If no incoming projection has the given target, the method returns zeros.

        :param target: the target.
        """
        return self.population.sum(target)[self.ranks]

    ################################
    ## Composition
    ################################

    def __add__(self, other):
        """Allows to join two PopulationViews if they have the same population."""
        from ANNarchy.core.Neuron import IndividualNeuron
        if other.population == self.population:
            if isinstance(other, IndividualNeuron):
                tmp = list(set(list(self.ranks) + [other.rank]))
                return PopulationView(self.population, np.array(tmp))
            elif isinstance(other, PopulationView):
                tmp = list(set(list(self.ranks) + list(other.ranks)))
                return PopulationView(self.population, np.array(tmp))
        else:
            Messages._error("can only add two PopulationViews of the same population.")

    def __repr__(self):
        """Defines the printing behaviour."""
        string ="PopulationView of " + str(self.population.name) + '\n'
        string += '  Ranks: ' +  str(self.ranks)
        string += '\n'
        for rk in self.ranks:
            string += '* ' + str(self.population.neuron(rk)) + '\n'
        return string



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
additional_global_functions = ""

init_launch_config = {
    'init': """
        // Generate the kernel launch configuration
        _threads_per_block = get_block_size();
        _nb_blocks = static_cast<unsigned int>(ceil(static_cast<double>(nb_dendrites())/static_cast<double>(_threads_per_block)));

    #ifdef _DEBUG
        std::cout << "Kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
        std::cerr << "re-configuring kernel config is not yet implemented for SELL" << std::endl;
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
""",
    'global': {
        'parameter':
    """
    // Global parameter %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global variable %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    bool %(name)s_host_to_device;
    long int %(name)s_device_to_host;
"""
    }
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = false;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = false;
        %(name)s_device_to_host = t;
""",
    'global': { 
        'parameter': """
        // Global parameter %(name)s
        %(name)s = 0.0;
""",
        'variable': """
        // Global variable %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), %(name)s.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), %(name)s.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, %(name)s.size() * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': "",
    'global': ""
}

#
#   Synaptic delays
#
delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        int delay""",
        'init': "delay = delays[0][0];",
        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    }
}

#
# Implement the continuous transmission for rate-code synapses.
#
rate_psp_kernel = {
    'device_kernel': {
        'sum':"""
__global__ void cu_proj%(id_proj)s_psp(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s) {
    //global row
    unsigned int row = blockIdx.x * blockDim.x + threadIdx.x;   

    if ( row < post_size ) {
        %(float_prec)s tmp = 0.0;
        unsigned int offset = row_ptr[blockIdx.x];
        unsigned int local_max_length = (row_ptr[blockIdx.x + 1] - offset)/block_size;
        
        for (unsigned int i = 0; i < local_max_length; i++) {
            %(idx_type)s idx = i * block_size + threadIdx.x + offset;
            %(idx_type)s rk_pre = rank_pre[idx];
            tmp += %(psp)s
        }
        %(target_arg)s%(post_index)s += tmp;
    }
}
"""
    },
    'invoke_kernel': """
void call_proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s) {
    cu_proj%(id_proj)s_psp<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* ranks and offsets */
        %(conn_args_call)s
        /* computation data */
        %(add_args_call)s
        /* result */
        %(target_arg_call)s
    );
}
""",
    'kernel_decl': """void call_proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        RunConfig proj%(id_proj)s_psp_cfg = RunConfig(proj%(id_proj)s->_nb_blocks, proj%(id_proj)s->_threads_per_block, 0, proj%(id_proj)s->stream);
        call_proj%(id_proj)s_psp (
            /* kernel config */
            proj%(id_proj)s_psp_cfg,
            /* ranks and offsets */
            %(conn_args)s
            /* computation data */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }    
}

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s post_size, const %(idx_type)s block_size, const %(size_type)s* __restrict__ row_ptr, const %(idx_type)s* __restrict__ rank_pre",
    'conn_call': "proj%(id_proj)s->nb_dendrites(), proj%(id_proj)s->get_block_size(), proj%(id_proj)s->d_row_ptr, proj%(id_proj)s->d_col_idx",
    'conn_kernel': "post_size, block_size, row_ptr, rank_pre",

    # launch config
    'launch_config': init_launch_config,
 
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,

    # operations
    'rate_psp': rate_psp_kernel,
    'synapse_update': {
        'global': None,
        'semiglobal': None,
        'local': None,
        'call': None
    }
}

conn_ids = {
    'local_index': "[idx]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[rk_pre]',
    'post_index': '[row]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_'
}


"""
Implements the built-in connectivity patterns available in ANNarchy. See the documentaton
for more details: https://annarchy.readthedocs.io/en/latest/manual/Connector.html

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""
import numpy as np

from ANNarchy.core.Random import RandomDistribution, DiscreteUniform
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.parser.report.LatexParser import _process_random
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

try:
    from ANNarchy.cython_ext import *
except Exception as e:
    Messages._print(e)

################################
## Connector methods
################################
def connect_one_to_one(self, 
                       weights: float | RandomDistribution = 1.0, 
                       delays: float | RandomDistribution = 0.0, 
                       force_multiple_weights:bool=False, 
                       storage_format:str=None, storage_order:str=None) -> "Projection":
    """
    one-to-one connection pattern.

    :param weights: Initial synaptic values, either a single value (float) or a random distribution object.
    :param delays: Synaptic delays, either a single value or a random distribution object (default=dt).
    :param force_multiple_weights: If a single value is provided for ``weights`` and there is no learning, a single weight value will be used for the whole projection instead of one per synapse. Setting ``force_multiple_weights`` to True ensures that a value per synapse will be used.
    """
    if self.pre.size != self.post.size:
        Messages._warning("connect_one_to_one() between", self.pre.name, 'and', self.post.name, 'with target', self.target)
        Messages._print("\t the two populations have different sizes, please check the connection pattern is what you expect.")

    self.connector_name = "One-to-One"
    self.connector_description = "One-to-One, weights %(weight)s, delays %(delay)s" % {'weight': _process_random(weights), 'delay': _process_random(delays)}

    if isinstance(weights, (int, float)) and not force_multiple_weights:
        self._single_constant_weight = True

    if storage_format == "dense":
        Messages._error("The usage of 'dense' storage format on one-to-one pattern is not allowed.")

    # if weights or delays are from random distribution I need to know this in code generator
    self.connector_weight_dist = weights if isinstance(weights, RandomDistribution) else None
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(one_to_one, (weights, delays, storage_format, storage_order), delays, storage_format, storage_order)

    return self

def connect_all_to_all(self, 
                       weights: float | RandomDistribution, 
                       delays: float | RandomDistribution =0.0, 
                       allow_self_connections:bool=False, 
                       force_multiple_weights:bool=False, 
                       storage_format:str=None, storage_order:str=None)  -> "Projection":
    """
    all-to-all (fully-connected) connection pattern.

    :param weights: Synaptic values, either a single value or a random distribution object.
    :param delays: Synaptic delays, either a single value or a random distribution object (default=dt).
    :param allow_self_connections: If True, self-connections between a neuron and itself are allowed (default = False if the pre- and post-populations are identical, True otherwise).
    :param force_multiple_weights: If a single value is provided for ``weights`` and there is no learning, a single weight value will be used for the whole projection instead of one per synapse. Setting ``force_multiple_weights`` to True ensures that a value per synapse will be used.
    """
    pre_pop = self.pre if not isinstance(self.pre, PopulationView) else self.pre.population
    post_pop = self.post if not isinstance(self.post, PopulationView) else self.post.population
    if pre_pop != post_pop:
        allow_self_connections = True

    self.connector_name = "All-to-All"
    self.connector_description = "All-to-All, weights %(weight)s, delays %(delay)s" % {'weight': _process_random(weights), 'delay': _process_random(delays)}

    # Does the projection define a single non-plastic weight?
    if isinstance(weights, (int, float)) and not force_multiple_weights:
        self._single_constant_weight = True

    # if weights or delays are from random distribution I need to know this in code generator
    self.connector_weight_dist = weights if isinstance(weights, RandomDistribution) else None
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    # Store the connectivity
    self._store_connectivity(all_to_all, (weights, delays, allow_self_connections, storage_format, storage_order), delays, storage_format, storage_order)

    return self

def connect_gaussian(self, amp:float, sigma:float, delays: float | RandomDistribution=0.0, limit:float=0.01, allow_self_connections:bool=False, storage_format:str=None)  -> "Projection":
    """
    Gaussian connection pattern.

    Each neuron in the postsynaptic population is connected to a region of the presynaptic population centered around the neuron with the same normalized coordinates using a Gaussian profile.

    :param amp: Amplitude of the Gaussian function
    :param sigma: Width of the Gaussian function
    :param delays: Synaptic delay, either a single value or a random distribution object (default=dt).
    :param limit: Proportion of `amp` below which synapses are not created
    :param allow_self_connections: Allows connections between a neuron and itself.
    """
    if self.pre != self.post:
        allow_self_connections = True

    if isinstance(self.pre, PopulationView) or isinstance(self.post, PopulationView):
        Messages._error('Gaussian connector is only possible on whole populations, not PopulationViews.')

    self.connector_name = "Gaussian"
    self.connector_description = f"Gaussian, amplitude {amp}, sigma {sigma}, delays {_process_random(delays)}"

    # weights are not drawn, delays possibly
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(gaussian, (amp, sigma, delays, limit, allow_self_connections, storage_format, "post_to_pre"), delays, storage_format, "post_to_pre")
    
    return self

def connect_dog(self, amp_pos:float, sigma_pos:float, amp_neg:float, sigma_neg:float, delays:float | RandomDistribution=0.0, limit:float=0.01, allow_self_connections:bool=False, storage_format:str=None)  -> "Projection":
    """
    Difference-Of-Gaussians connection pattern.

    Each neuron in the postsynaptic population is connected to a region of the presynaptic population centered around the neuron with the same normalized coordinates using a Difference-Of-Gaussians profile.

    :param amp_pos: Amplitude of the positive Gaussian function
    :param sigma_pos: Width of the positive Gaussian function
    :param amp_neg: Amplitude of the negative Gaussian function
    :param sigma_neg: Width of the negative Gaussian function
    :param delays: Synaptic delay, either a single value or a random distribution object (default=dt).
    :param limit: Proportion of *amp* below which synapses are not created (default: 0.01)
    :param allow_self_connections: Allows connections between a neuron and itself.
    """
    if self.pre != self.post:
        allow_self_connections = True

    if isinstance(self.pre, PopulationView) or isinstance(self.post, PopulationView):
        Messages._error('DoG connector is only possible on whole populations, not PopulationViews.')

    self.connector_name = "Difference-of-Gaussian"
    self.connector_description = f"Difference-of-Gaussian, A+ {amp_pos}, sigma+ {sigma_pos}, A- {amp_neg}, sigma- {sigma_neg}, delays {_process_random(delays)}"

    # delays are possibly drawn from distribution, weights not
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(dog, (amp_pos, sigma_pos, amp_neg, sigma_neg, delays, limit, allow_self_connections, storage_format, "post_to_pre"), delays, storage_format, "post_to_pre")

    return self

def connect_fixed_probability(self, probability:float, weights:float | RandomDistribution, delays:float | RandomDistribution=0.0, allow_self_connections:bool=False, force_multiple_weights:bool=False, storage_format:str=None, storage_order:str=None)  -> "Projection":
    """
    Probabilistic sparse connection pattern.

    Each neuron in the postsynaptic population is connected to neurons of the presynaptic population with the given probability. Self-connections are avoided by default.

    :param probability: Probability that a synapse is created.
    :param weights: Either a single value for all synapses or a RandomDistribution object.
    :param delays: Either a single value for all synapses or a RandomDistribution object (default = dt)
    :param allow_self_connections: Defines if self-connections are allowed (default=False).
    :param force_multiple_weights: If a single value is provided for `weights` and there is no learning, a single weight value will be used for the whole projection instead of one per synapse. Setting `force_multiple_weights` to True ensures that a value per synapse will be used.
    """
    if self.pre != self.post:
        allow_self_connections = True

    self.connector_name = "Random"
    self.connector_description = "Random, sparseness %(proba)s, weights %(weight)s, delays %(delay)s" % {'weight': _process_random(weights), 'delay': _process_random(delays), 'proba': probability}

    if isinstance(weights, (int, float)) and not force_multiple_weights:
        self._single_constant_weight = True

    # if weights or delays are from random distribution I need to know this in code generator
    self.connector_weight_dist = weights if isinstance(weights, RandomDistribution) else None
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(fixed_probability, (probability, weights, delays, allow_self_connections, storage_format, storage_order), delays, storage_format, storage_order)

    return self

def connect_fixed_number_pre(self, number:int, weights: float | RandomDistribution, delays: float | RandomDistribution=0.0, allow_self_connections:bool=False, force_multiple_weights:bool=False, storage_format:str=None, storage_order:str=None)  -> "Projection":
    """
    Connection pattern where each post-synaptic neuron receives a fixed number of pre-synaptic neurons.

    :param number: Number of synapses per postsynaptic neuron.
    :param weights: Either a single value for all synapses or a RandomDistribution object.
    :param delays: Either a single value for all synapses or a RandomDistribution object (default = dt)
    :param allow_self_connections: Defines if self-connections are allowed (default=False).
    :param force_multiple_weights: If a single value is provided for ``weights`` and there is no learning, a single weight value will be used for the whole projection instead of one per synapse. Setting ``force_multiple_weights`` to True ensures that a value per synapse will be used.
    """
    if self.pre != self.post:
        allow_self_connections = True

    if number > self.pre.size:
        Messages._error('fixed_number_pre: the number of pre-synaptic neurons exceeds the size of the population.')

    self.connector_name = "Random Convergent"
    self.connector_description = "Random Convergent %(number)s $\\rightarrow$ 1, weights %(weight)s, delays %(delay)s"% {'weight': _process_random(weights), 'delay': _process_random(delays), 'number': number}

    if isinstance(weights, (int, float)) and not force_multiple_weights:
        self._single_constant_weight = True

    # if weights or delays are from random distribution I need to know this in code generator
    self.connector_weight_dist = weights if isinstance(weights, RandomDistribution) else None
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(fixed_number_pre, (number, weights, delays, allow_self_connections, storage_format, storage_order), delays, storage_format, storage_order)

    return self

def connect_fixed_number_post(self, number:int, weights:float | RandomDistribution=1.0, delays:float | RandomDistribution=0.0, allow_self_connections:bool=False, force_multiple_weights:bool=False, storage_format:str=None, storage_order:str=None)  -> "Projection":
    """
    Each pre-synaptic neuron randomly sends a fixed number of connections to the post-synaptic neurons.

    :param number: Number of synapses per pre-synaptic neuron.
    :param weights: Either a single value for all synapses or a RandomDistribution object.
    :param delays: Either a single value for all synapses or a RandomDistribution object (default = dt)
    :param allow_self_connections: Defines if self-connections are allowed (default=False)
    :param force_multiple_weights: If a single value is provided for ``weights`` and there is no learning, a single weight value will be used for the whole projection instead of one per synapse. Setting ``force_multiple_weights`` to True ensures that a value per synapse will be used.
    """
    if self.pre != self.post:
        allow_self_connections = True

    if number > self.post.size:
        Messages._error('fixed_number_post: the number of post-synaptic neurons exceeds the size of the population.')

    self.connector_name = "Random Divergent"
    self.connector_description = "Random Divergent 1 $\\rightarrow$ %(number)s, weights %(weight)s, delays %(delay)s"% {'weight': _process_random(weights), 'delay': _process_random(delays), 'number': number}

    if isinstance(weights, (int, float)) and not force_multiple_weights:
        self._single_constant_weight = True

    # if weights or delays are from random distribution I need to know this in code generator
    self.connector_weight_dist = weights if isinstance(weights, RandomDistribution) else None
    self.connector_delay_dist = delays if isinstance(delays, RandomDistribution) else None

    self._store_connectivity(fixed_number_post, (number, weights, delays, allow_self_connections, storage_format, storage_order), delays, storage_format, storage_order)
    return self


def connect_with_func(self, method, storage_format:str=None, storage_order:str=None, **args) -> "Projection":
    """
    Connection pattern based on a user-defined function.

    The two first arguments of the function must be the pre and post populations. Additional arguments can be passed at creation time.

    The function must return a `ann.LILConnectivity` object.

    Example:

    ```python
    def probabilistic_pattern(pre, post, weight, probability):
        # Create a LIL structure for the connectivity matrix
        synapses = ann.LILConnectivity()
        # For all neurons in the post-synaptic population
        for post_rank in xrange(post.size):
            # Decide which pre-synaptic neurons should form synapses
            ranks = []
            for pre_rank in xrange(pre.size):
                if random.random() < probability:
                    ranks.append(pre_rank)
            # Create weights and delays arrays of the same size
            values = [weight for i in xrange(len(ranks)) ]
            delays = [0 for i in xrange(len(ranks)) ]
            # Add this information to the LIL matrix
            synapses.add(post_rank, ranks, values, delays)

        return synapses

    proj = ann.Projection(pop1, pop2, target = 'inh')
    proj.connect_with_func(
        method=probabilistic_pattern, 
        weight=1.0, 
        probability=0.3
    ) 

    ```

    :param method: Method to call. The method **must** return a LILConnectivity object.
    :param args: List of additional arguments needed by the function.
    """
    # Invoke the method directly, we need the delays already....
    synapses = method(self.pre, self.post, **args)
    synapses.validate()

    # Treat delays
    if synapses.uniform_delay != -1: # uniform delay
        d = synapses.max_delay * ConfigManager().get('dt', self.net_id)
        self.connector_delay_dist = None
    else:
        # Just to trick _store_connectivity(), the real delays are in the CSR
        d = DiscreteUniform(0., synapses.max_delay * ConfigManager().get('dt', self.net_id))
        self.connector_delay_dist = DiscreteUniform(0., synapses.max_delay * ConfigManager().get('dt', self.net_id))

    self._store_connectivity(self._load_from_lil, (synapses, ), d, storage_format=storage_format, storage_order=storage_order)

    self.connector_name = "User-defined"
    self.connector_description = "Created by the method " + method.__name__
    return self

def connect_from_matrix_market(self, filename:str, storage_format:str=None, storage_order:str=None) -> "Projection":
    """
    Loads a weight matrix encoded in the Matrix Market format. This connector is intended for benchmarking purposes.

    :param filename: Filename of the Matrix Market (.mtx) file.
    """
    from scipy.io import mmread
    from scipy.sparse import coo_matrix
    import tarfile

    from ANNarchy.cython_ext import LILConnectivity
    if not filename.endswith(".mtx"):
        raise ValueError("from_matrix_market(): expected .mtx file.")

    # read with SciPy
    tmp = mmread(filename)

    # scipy should return a coo_matrix in case of sparse matrices
    if isinstance(tmp, coo_matrix):
        # transform into LIL (in place)
        tmp = tmp.tolil(copy=True)

        # build up ANNarchy LIL
        synapses = LILConnectivity()
        row_idx = 0
        for col_idx, val in zip(tmp.rows, tmp.data):
            synapses.push_back(row_idx, col_idx, val, [0])
            row_idx+=1

    elif isinstance(tmp, np.ndarray):
        # build up ANNarchy LIL
        synapses = LILConnectivity()

        col_idx = np.arange(tmp.shape[1])
        for row_idx in range(tmp.shape[0]):
            synapses.push_back(row_idx, col_idx, tmp[row_idx,:], [0])

    else:
        raise ValueError("Error on read-out of matrix market file.")

    # not needed anymore
    del tmp

    delays = 0

    self._store_connectivity(self._load_from_lil, (synapses, ), delays, storage_format=storage_format, storage_order=storage_order)

    self.connector_name = "MatrixMarket"
    self.connector_description = "A weight matrix load from .mtx file"
    return self

def _load_from_lil(self, pre, post, synapses):
    """
    Load from LILConnectivity instance.
    """
    return synapses

def connect_from_matrix(self, weights: np.array, delays=0.0, pre_post=False, storage_format=None, storage_order=None) -> "Projection":
    """
    Builds a connection pattern according to a dense connectivity matrix.

    The matrix must be N*M, where N is the number of neurons in the post-synaptic population and M in the pre-synaptic one. Lists of lists must have the same size.

    If a synapse should not be created, the weight value should be None.

    :param weights: Numpy array (or list of lists of equal size) representing the weights. If a value is None, the corresponding synapse will not be created.
    :param delays: Numpy array representing the delays. Must represent the same synapses as the `weights` argument. If omitted, the delays are considered 0.
    :param pre_post: States which index is first. By default, the first dimension is related to the post-synaptic population. If ``pre_post`` is True, the first dimension is the pre-synaptic population.
    """

    # Store the synapses
    self.connector_name = "Connectivity matrix"
    self.connector_description = "Connectivity matrix"

    if isinstance(weights, list):
        try:
            weights = np.array(weights)
        except:
            Messages._error('from_matrix(): You must provide a dense 2D matrix.')

    self._store_connectivity(self._load_from_matrix, (weights, delays, pre_post), delays, storage_format, storage_order)

    return self

def _load_from_matrix(self, pre, post, weights, delays, pre_post):
    """
    Initializes a connectivity matrix between two populations based on a provided matrix.
    This connector method always return a LIL-structure.

    :param pre: pre-synaptic Population instance
    :param post: post-synaptic Population instance
    :param weights: matrix / list-in-list which contains synaptic weights
    :param delays: matrix / list-in-list which contains synaptic delays
    :param pre_post: needs to be set to True if the weights are not a post times pre matrix.
    """
    lil = LILConnectivity()

    uniform_delay = not isinstance(delays, (list, np.ndarray))
    if isinstance(delays, list):
        try:
            delays = np.array(delays)
        except:
            Messages._error('from_matrix(): You must provide a dense 2D matrix.')

    if pre_post: # if the user prefers pre as the first index...
        weights = weights.T
        if isinstance(delays, np.ndarray):
            delays = delays.T

    shape = weights.shape
    if shape != (self.post.size, self.pre.size):
        if not pre_post:
            Messages._print("ERROR: from_matrix(): the matrix does not have the correct dimensions.")
            Messages._print('Expected:', (self.post.size, self.pre.size))
            Messages._print('Received:', shape)

        else:
            Messages._print("ERROR: from_matrix(): the matrix does not have the correct dimensions.")
            Messages._print('Expected:', (self.pre.size, self.post.size))
            Messages._print('Received:', shape)
        Messages._error('Quitting...')

    for i in range(self.post.size):
        if isinstance(self.post, PopulationView):
            rk_post = self.post.ranks[i]
        else:
            rk_post = i
        r = []
        w = []
        d = []
        for j in range(self.pre.size):
            val = weights[i, j]
            if val != None:
                if isinstance(self.pre, PopulationView):
                    rk_pre = self.pre.ranks[j]
                else:
                    rk_pre = j
                r.append(rk_pre)
                w.append(val)
                if not uniform_delay:
                    d.append(delays[i, j])
        if uniform_delay:
            d.append(delays)
        if len(r) > 0:
            lil.add(rk_post, r, w, d)

    return lil

def connect_from_sparse(self, weights:"scipy.sparse.lil_matrix", delays: int | float=0.0, storage_format:str=None, storage_order:str=None) -> "Projection":
    """
    Builds a connectivity pattern using a Scipy sparse matrix for the weights and (optionally) delays.

    Warning: a sparse matrix has pre-synaptic ranks as first dimension.

    :param weights: a sparse lil_matrix object created from scipy.
    :param delays: the value of the constant delay (default: dt). Variable delays are not allowed.
    """
    try:
        from scipy.sparse import lil_matrix, csr_matrix, csc_matrix
    except:
        Messages._error("from_sparse(): scipy is not installed, sparse matrices can not be loaded.")

    if not isinstance(weights, (lil_matrix, csr_matrix, csc_matrix)):
        Messages._error("from_sparse(): only lil, csr and csc matrices are allowed for now.")

    if not isinstance(delays, (int, float)):
        Messages._error("from_sparse(): only constant delays are allowed for sparse matrices.")

    weights = csc_matrix(weights)

    # if weights[weights.nonzero()].max() == weights[weights.nonzero()].min() :
    #     self._single_constant_weight = True

    # Store the synapses
    self.connector_name = "Sparse connectivity matrix"
    self.connector_description = "Sparse connectivity matrix"
    self._store_connectivity(self._load_from_sparse, (weights, delays), delays, storage_format, storage_order)

    return self

def _load_from_sparse(self, pre, post, weights, delays):
    # Create an empty LIL object
    lil = LILConnectivity()

    # Find offsets
    if isinstance(self.pre, PopulationView):
        pre_ranks = self.pre.ranks
    else:
        pre_ranks = [i for i in range(self.pre.size)]

    if isinstance(self.post, PopulationView):
        post_ranks = self.post.ranks
    else:
        post_ranks = [i for i in range(self.post.size)]

    # Process the sparse matrix and fill the lil
    weights.sort_indices()
    (pre, post) = weights.shape

    if (pre, post) != (len(pre_ranks), len(post_ranks)):
        Messages._print("ERROR: from_sparse(): the sparse matrix does not have the correct dimensions.")
        Messages._print('Expected:', (len(pre_ranks), len(post_ranks)))
        Messages._print('Received:', (pre, post))
        Messages._error('Quitting...')

    for idx_post in range(post):
        idx_pre = weights.getcol(idx_post).indices
        w = weights.getcol(idx_post).data
        pr = [pre_ranks[i] for i in idx_pre]
        lil.add(post_ranks[idx_post], pr, w, [float(delays)])

    return lil

def connect_from_file(self, filename:str, pickle_encoding:str=None, storage_format:str=None, storage_order:str=None)  -> "Projection":
    """
    Builds the connectivity matrix using data saved using `Projection.save_connectivity()` (not `save()`!).

    Admissible file formats are compressed Numpy files (.npz), gunzipped binary text files (.gz) or binary text files.

    Note: Only the ranks, weights and delays are loaded, not the other variables.

    :param filename: file where the connections were saved.
    """
    # Create an empty LIL object
    lil = LILConnectivity()

    # Load the data
    from ANNarchy.core.IO import _load_connectivity_data
    try:
        data = _load_connectivity_data(filename, pickle_encoding)
    except Exception as e:
        Messages._print(e)
        Messages._error('from_file(): Unable to load the data', filename, 'into the projection.')

    # Load the LIL object
    try:
        # Size
        lil.size = data['size']
        lil.nb_synapses = data['nb_synapses']

        # Ranks
        lil.post_rank = list(data['post_ranks'])
        lil.pre_rank = list(data['pre_ranks'])

        # Weights
        if isinstance(data['w'], (int, float)):
            self._single_constant_weight = True
            lil.w = [[float(data['w'])]]
        elif isinstance(data['w'], (np.ndarray,)) and data['w'].size == 1:
            self._single_constant_weight = True
            lil.w = [[float(data['w'])]]
        else:
            lil.w = data['w']

        # Delays
        lil.max_delay = data['max_delay']
        lil.uniform_delay = data['uniform_delay']

        if data['delay'] is not None:
            if lil.uniform_delay == -1:
                lil.delay = list(data['delay'])
            else:
                lil.delay = [[lil.max_delay]]

    except Exception as e:
        Messages._print(e)
        Messages._error('Unable to load the data', filename, 'into the projection.')

    # Store the synapses
    self.connector_name = "From File"
    self.connector_description = "From File"
    self._store_connectivity(self._load_from_lil, (lil,), lil.max_delay if lil.uniform_delay > 0 else lil.delay, storage_format=storage_format, storage_order=storage_order)

    return self



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from .ParserTemplate import create_local_dict, user_functions

import sympy as sp
import re
from copy import deepcopy

def transform_condition(expr):
    """
    Transforms the "natural" logical operators into Sympy-compatible versions.
    """
    expr = expr.replace (' and ', ' & ')
    expr = expr.replace (' or ', ' | ')
    expr = expr.replace (' is not ', ' != ')
    expr = expr.replace (' not ', ' Not ')
    expr = expr.replace (' not(', ' Not(')
    expr = expr.replace (' is ', ' == ')

    return expr

class Equation(object):
    '''
    Class to analyse a single equation.
    '''
    def __init__(self,
                 name,
                 expression,
                 description,
                 untouched = [],
                 method='explicit',
                 type=None,
                 net_id=0,
        ):
        '''
        Parameters:

        * name : The name of the variable
        * expression: The expression as a string
        * variables: a list of all the variables in the neuron/synapse
        * local_variables: a list of the local variables
        * global_variables: a list of the global variables
        * method: the numerical method to use for ODEs
        * type: forces the analyser to consider the equation as: simple, cond, ODE, inc, otherwise it is guessed.
        * untouched: list of terms which should not be modified
        '''

        # Store attributes
        self.name = name
        self.expression = transform_condition(expression)
        self.description = description
        self.attributes = self.description['attributes']
        self.local_attributes = self.description['local']
        self.semiglobal_attributes = self.description['semiglobal']
        self.global_attributes = self.description['global']
        self.local_functions = [func['name'] for func in self.description['functions']]
        self.variables = [var['name'] for var in self.description['variables']]
        self.untouched = untouched
        self.method = method
        self.num_flops = 0
        self.net_id=net_id

        # Determine the type of the equation
        if not type:
            self.type = self.identify_type()
        else:
            self.type = type

        # Copy the list of built-in functions
        self.user_functions = user_functions.copy()

        # Add each user-defined function to avoid "not supported in C"
        for var in self.local_functions:
            self.user_functions[var] = var

        # Copy the default dictionary of built-in symbols or functions
        self.local_dict = create_local_dict(
            self.local_attributes,
            self.semiglobal_attributes,
            self.global_attributes,
            self.untouched,
            self.user_functions
            ) # in ParserTemplate


    def parse(self):
        "Main method called after creating the object."
        try:
            if self.type == 'ODE':
                code = self.analyse_ODE(self.expression)
            elif self.type == 'cond':
                code = self.analyse_condition(self.expression)
            elif self.type == 'inc':
                code = self.analyse_increment(self.expression)
            elif self.type == 'return':
                code = self.analyse_return(self.expression)
            elif self.type == 'simple':
                code = self.analyse_assignment(self.expression)
        except Exception as e:
            Messages._print(e)
            Messages._error('Parser: cannot analyse', self.expression)
        return code

    def identify_type(self):
        """
        Identifies which type has the equation:

        * inc for "a += 0.2"
        * ODE for "dV/dt + V = A"
        * cond for "mp > 30.0" or "r != 0.0"
        * simple for the rest, e.g. "r = pos(mp)" or "baseline = Uniform(0,1)"
        """
        # Suppress spaces to extract dvar/dt
        expression = self.expression.replace(' ', '')

        # Check if it an increment
        for op in ['+=', '-=', '*=', '/=']:
            if op in expression:
                return 'inc'

        # Check if it is an ode
        if 'd'+self.name+'/dt' in expression:
            return 'ODE'

        # Check if it is a condition (e.g spike)
        split_expr = expression.split('=')
        if len(split_expr) == 1: # no equal sign = treat as condition (return statements will be specified
            return 'cond'
        if split_expr[1] == '': # the first = is followed by another =
            return 'cond'
        for logop in ['<', '>', '!']: # the first operator is <=, >= or !=
            if logop == split_expr[0][-1]:
                return 'cond'

        # Only a simple assignement
        return 'simple'


    ###############################################
    ### Code generation
    ###############################################

    def c_code(self, equation):
        """
        Returns the C version of a Sympy expression

        Implementation note:

        Changes to this method should be applied also
        on ANNarchy.parser.CoupledEquations.c_code() too.
        """
        # The `strict` parameter has been introduced in sympy >= 1.13.
        # It defaults to `True`.
        try:
            c_code = sp.ccode(
                equation,
                precision=8,
                strict=False,
                user_functions=self.user_functions
            )
        except TypeError:
            c_code = sp.ccode(
                equation,
                precision=8,
                user_functions=self.user_functions
            )

        if ConfigManager().get('precision', self.net_id) == "float":
            #
            # Add the f-suffix to floating value constants
            matches = re.findall(r"[-]?[0-9]+\.[0-9]+", c_code)
            matches = sorted(list(set(matches)))    # remove doublons, e. g. 0.5*dt

            for m in matches:
                fval = float(m)
                fval = round(fval, 8)       # shorten the val to a reasonable length
                c_code = c_code.replace(m, str(fval)+"f")

            # If found constants have an overlap, e.g., 5.0 and 15.0, we create
            # multiple suffixes, e.g., 15.0ff. Therefore, we need to remove them
            # afterwards ...
            matches = re.findall(r"[-]?[0-9]+\.[0-9]+f[f]+", c_code)
            for m in matches:
                fval = float(m.split("f")[0])
                fval = round(fval, 8)       # shorten the val to a reasonable length
                c_code = c_code.replace(m, str(fval)+"f")

            # Replace the math functions with their single precision
            # to circumenvent problems induced by implicit type conversion
            # e. g. pow(double, double) by powf(float, float)
            for func in ["fabs", "pow", "exp", "sin", "cos", "tan", "round"]:
                c_code = c_code.replace(func+"(", func+"f(")

        return c_code

    def latex_code(self, equation):
        "Returns the LaTeX version of a Sympy expression"
        return sp.latex(equation)

    def parse_expression(self, expression, local_dict):
        " Parses a string with respect to the vocabulary defined in local_dict."
        from sympy.parsing.sympy_parser import parse_expr, standard_transformations, convert_xor, auto_number, rationalize
        try:
            res =  parse_expr(
                expression,
                local_dict = local_dict,
                transformations = (standard_transformations + (convert_xor, )),
                evaluate=False
            )
        except Exception as e:
            Messages._print(e)
            Messages._error('Can not analyse the expression :' +  str(expression))

        return res

    def _count_ops(self, equation):
        """
        Count the number of numeric operations to approximate the number of required
        floating point operations.
        """
        exp_symbol = sp.Symbol('EXP')

        # Some numeric operations are more expensive than others. For example,
        # Izhikevich (2004) states ten operations for exponential functions
        weighted_count = sp.count_ops(equation, visual=True).subs(exp_symbol, 10)

        # Every other operation gets a weight of 1 (the default)
        weighted_count = weighted_count.replace(sp.Symbol, type(sp.S.One))

        # return the (re-)weighted result
        return weighted_count

    ###############################################
    ### ODE
    ###############################################

    def analyse_ODE(self, expression):
        " Returns the C++ code corresponding to an ODE with the method defined in self.method"
        # transform the expression to suppress =
        if '=' in expression:
            expression = expression.replace('=', '- (')
            expression += ')'

        # Suppress spaces to extract dvar/dt
        expression = expression.replace(' ', '')

        if self.method == 'semiimplicit':
            return self.semiimplicit(expression)
        elif self.method == 'implicit':
            return self.implicit(expression)
        elif self.method == 'explicit':
            return self.explicit(expression)
        elif self.method == 'exponential':
            return self.exponential(expression)
        elif self.method == 'midpoint':
            return self.midpoint(expression)
        elif self.method == 'rk4':
            return self.runge_kutta_4(expression)
        elif self.method == 'event-driven':
            return self.eventdriven(expression)
        else:
            raise NotImplementedError("The chosen numerical method ", self.method, "is not available")

    def explicit(self, expression):
        " Explicit or backward Euler numerical method"

        expression = expression.replace('d'+self.name+'/dt', '_grad_var_')
        new_var = sp.Symbol('_grad_var_')
        self.local_dict['_grad_var_'] = new_var

        analysed = self.parse_expression(expression,
            local_dict = self.local_dict
        )

        self.analysed = analysed
        variable_name = self.local_dict[self.name]

        equation = sp.solve(analysed, new_var, check=False, rational=False)[0]
        equation = sp.simplify(equation, ratio=1.0)

        explicit_code = ConfigManager().get('precision', self.net_id) + ' _' + self.name + ' = ' + self.c_code(equation) + ';'

        switch = self.c_code(variable_name) + ' += dt*_' + self.name + ' ;'

        # compute required number of operations
        self.num_flops = self._count_ops(equation)    # compute increment
        self.num_flops += 2                          # apply increment

        # Return result
        return [{}, explicit_code, switch]


    def midpoint(self, expression):

        "Midpoint method."
        expression = expression.replace('d'+self.name+'/dt', '_grad_var_')
        new_var = sp.Symbol('_grad_var_')
        self.local_dict['_grad_var_'] = new_var

        analysed = self.parse_expression(expression,
            local_dict = self.local_dict
        )

        self.analysed = analysed

        variable_name = self.local_dict[self.name]

        # equation = sp.simplify(sp.collect( sp.solve(analysed, new_var, check=False)[0], self.local_dict['dt']))
        equation = sp.solve(analysed, new_var, check=False, rational=False)[0]
        equation = sp.collect(equation, self.local_dict['dt'])
        equation = sp.simplify(equation, ratio=1.0)

        explicit_code = ConfigManager().get('precision', self.net_id) + ' _k_' + self.name + ' = dt*(' + self.c_code(equation) + ');'
        # Midpoint method:
        # Replace the variable x by x+_x/2
        tmp_dict = self.local_dict
        tmp_dict[self.name] = sp.Symbol('(' + self.c_code(variable_name) + ' + 0.5*_k_' + self.name + ' )')
        tmp_analysed = self.parse_expression(expression,
            local_dict = self.local_dict
        )
        tmp_equation = sp.solve(tmp_analysed, new_var, check=False, rational=False)[0]
        explicit_code += '\n' + ConfigManager().get('precision', self.net_id) + ' _' + self.name + ' = ' + self.c_code(tmp_equation) + ';'

        switch = self.c_code(variable_name) + ' += dt*_' + self.name + ' ;'

        # compute required number of operations
        self.num_flops = self._count_ops(equation)        # compute increment (explicit)
        self.num_flops += self._count_ops(tmp_equation)   # compute increment (dt/2.0)
        self.num_flops += 2                              # apply increment

        # Return result
        return [{}, explicit_code, switch]

    def runge_kutta_4(self, expression):
        "Runge-Kutta 4th order method"

        expression = expression.replace('d'+self.name+'/dt', '_grad_var_')
        new_var = sp.Symbol('_grad_var_')
        self.local_dict['_grad_var_'] = new_var

        analysed = self.parse_expression(expression,
            local_dict = self.local_dict
        )

        self.analysed = analysed

        variable_name = self.local_dict[self.name]

        # k1 = f(x)
        equation = sp.solve(analysed, new_var, check=False, rational=False)[0]
        equation = sp.collect(equation, self.local_dict['dt'])
        equation = sp.simplify(equation, ratio=1.0)
        explicit_code = ConfigManager().get('precision', self.net_id) + ' _k1_' + self.name + ' = (' + self.c_code(equation) + ');\n'

        # k2 = f(x+dt/2*k)
        tmp_dict = deepcopy(self.local_dict)
        tmp_dict[self.name] = sp.Symbol('(' + self.c_code(variable_name) + ' + 0.5 * dt * _k1_' + self.name + ' )')
        tmp_analysed = self.parse_expression(expression,
            local_dict = tmp_dict
        )
        tmp_equation_k2 = sp.solve(tmp_analysed, new_var, check=False, rational=False)[0]
        explicit_code += ConfigManager().get('precision', self.net_id) + ' _k2_' + self.name + ' = (' + self.c_code(tmp_equation_k2) + ');\n'

        # k3 = f(x+dt/2*k2)
        tmp_dict = deepcopy(self.local_dict)
        tmp_dict[self.name] = sp.Symbol('(' + self.c_code(variable_name) + ' + 0.5 * dt * _k2_' + self.name + ' )')
        tmp_analysed = self.parse_expression(expression,
            local_dict = tmp_dict
        )
        tmp_equation_k3 = sp.solve(tmp_analysed, new_var, check=False, rational=False)[0]
        explicit_code += ConfigManager().get('precision', self.net_id) + ' _k3_' + self.name + ' = (' + self.c_code(tmp_equation_k3) + ');\n'

        # k4 = f(x+dt*k3)
        tmp_dict = deepcopy(self.local_dict)
        tmp_dict[self.name] = sp.Symbol('(' + self.c_code(variable_name) + ' + dt * _k3_' + self.name + ' )')
        tmp_analysed = self.parse_expression(expression,
            local_dict = tmp_dict
        )
        tmp_equation_k4 = sp.solve(tmp_analysed, new_var, check=False, rational=False)[0]
        explicit_code += ConfigManager().get('precision', self.net_id) + ' _k4_' + self.name + ' = (' + self.c_code(tmp_equation_k4) + ');\n'

        # final x is part of k1 .. k4
        switch = self.c_code(variable_name) + ' += dt/6.0 * ( _k1_' + self.name + ' + (_k2_' + self.name + '+_k2_' + self.name + ') + (_k3_' + self.name + '+_k3_' + self.name + ') + _k4_' + self.name + ');'

        # compute required number of operations
        self.num_flops = self._count_ops(equation)            # compute increment (explicit)
        self.num_flops += self._count_ops(tmp_equation_k2)    # compute increment (k2)
        self.num_flops += self._count_ops(tmp_equation_k3)    # compute increment (k3)
        self.num_flops += self._count_ops(tmp_equation_k4)    # compute increment (k4)
        self.num_flops += 8                                   # apply increment

        return [{}, explicit_code, switch]

    def implicit(self, expression):
        "Full implicit method, linearising for example (V - E)^2, but this is not desired."

        # print('Expression', expression)
        # Transform the gradient into a difference TODO: more robust...
        new_expression = expression.replace('d'+self.name, '_t_gradient_')
        new_expression = re.sub(r'([^\w]+)'+self.name+r'([^\w]+)', r'\1_'+self.name+r'\2', new_expression)
        new_expression = new_expression.replace('_t_gradient_', '(_'+self.name+' - '+self.name+')')

        # print('New Expression', new_expression)

        # Add a sympbol for the next value of the variable
        new_var = sp.Symbol('_'+self.name)
        self.local_dict['_'+self.name] = new_var

        # Parse the string
        analysed = self.parse_expression(
            new_expression,
            local_dict = self.local_dict
        )
        self.analysed = analysed
        # print('Analysed', analysed)

        # Solve the equation for delta_mp
        solved = sp.solve(analysed, new_var, check=False, rational=False)
        # print('Solved', solved)
        if len(solved) > 1:
            Messages._print(self.expression)
            Messages._error('Parser: the ODE is not linear, can not use the implicit method.')

        else:
            solved = solved[0]

        equation = sp.simplify(sp.collect( solved, self.local_dict['dt']))

        # Obtain C code
        variable_name = self.c_code(self.local_dict[self.name])

        explicit_code = ConfigManager().get('precision', self.net_id) + ' _' + self.name + ' = '\
                        +  self.c_code(equation) + ';'
        switch = variable_name + ' = _' + self.name + ' ;'

        # Return result
        return [{}, explicit_code, switch]

    def semiimplicit(self, expression):
        " Implicit or forward Euler numerical method, but only for the linear part of the equation."
        # Standardize the equation
        real_tau, stepsize, steadystate = self.standardize_ODE(expression)

        if real_tau is None: # the equation can not be standardized
            Messages._print(expression)
            Messages._warning('The implicit Euler method can not be applied to this equation (must be linear), applying explicit Euler instead.')
            return self.explicit(expression)

        instepsize = sp.together( stepsize / (stepsize + sp.S(1.0)) )

        # Obtain C code
        variable_name = self.c_code(self.local_dict[self.name])

        explicit_code = ConfigManager().get('precision', self.net_id) + ' _' + self.name + ' = ('\
                        +  self.c_code(instepsize) + ')*(' \
                        + self.c_code(steadystate)+ ' - ' + variable_name +');'
        switch = variable_name + ' += _' + self.name + ' ;'

        # Return result
        return [{}, explicit_code, switch]


    def exponential(self, expression):
        "Exponential Euler method."

        # Standardize the equation
        real_tau, stepsize, steadystate = self.standardize_ODE(expression)

        if real_tau is None: # the equation can not be standardized
            return self.explicit(expression)

        # Check if the step size is local or not
        global_stepsize = True
        for var in self.attributes: # Add each variable of the neuron
            if var in self.local_attributes and self.local_dict[var] in stepsize.atoms():
                global_stepsize = False

        # If the stepsize is global, take it out of the for loop to save some operations
        pre_loop = {}
        step_size_code = '(1.0 - exp(' + self.c_code(-stepsize) + '))'
        if global_stepsize:
            pre_loop['name'] = "__stepsize_" + self.name
            pre_loop['value'] = '1.0 - exp( ' + self.c_code(-stepsize) + ')'
            step_size_code = "__stepsize_" + self.name


        # Obtain C code
        variable_name = self.c_code(self.local_dict[self.name])

        explicit_code = ConfigManager().get('precision', self.net_id) + ' _' + self.name + ' =  ' \
                        + step_size_code \
                        + '*(' \
                        + self.c_code(steadystate) \
                        + ' - ' \
                        + self.c_code(self.local_dict[self.name]) \
                        +');'

        switch = variable_name + ' += _' + self.name + ' ;'

        # Return result
        return [pre_loop, explicit_code, switch]

    def eventdriven(self, expression):
        # Standardize the equation
        real_tau, _, steadystate = self.standardize_ODE(expression)

        if real_tau is None: # the equation can not be standardized
            Messages._print(self.expression)
            Messages._error('The equation is not a linear ODE and can not be evaluated exactly.')


        # Check the steady state is not dependent on other variables
        for var in self.variables:
            if self.local_dict[var] in steadystate.atoms():
                Messages._print(self.expression)
                Messages._error('The equation can not depend on other variables ('+var+') to be evaluated exactly.')


        # Obtain C code
        variable_name = self.c_code(self.local_dict[self.name])
        steady = self.c_code(steadystate)
        if steady == '0':
            code = variable_name + ' *= exp(dt*(_last_event%(local_index)s - (t))/(' + self.c_code(real_tau) + '));'
        else:
            code = variable_name + ' = ' + steady + ' + (' + variable_name + ' - ' + steady + ')*exp(dt*(_last_event%(local_index)s - (t))/(' + self.c_code(real_tau) + '));'
        return code

    def standardize_ODE(self, expression):
        """ Transform any 1rst order ODE into the standardized form:

        tau * dV/dt + V = S

        Non-linear functions of V are left in the steady-state argument.

        Returns:

            * tau : the time constant associated to the standardized equation.

            * stepsize: a simplified version of dt/tau.

            * steadystate: the right term of the equation after standardization
        """
        # Replace the gradient with a temporary variable
        expression = expression.replace('d' + self.name +'/dt', '_gradvar_') # TODO: robust to spaces

        # Add the gradient sympbol
        grad_var = sp.Symbol('_gradvar_')

        # Parse the string
        analysed = self.parse_expression(
            expression,
            local_dict = self.local_dict
        )
        self.analysed = analysed

        # Collect factor on the gradient and main variable A*dV/dt + B*V = C
        expanded = analysed.expand(
            modulus=None, power_base=False, power_exp=False,
            mul=True, log=False, multinomial=False)

        # Make sure the expansion went well
        collected_var = sp.collect(expanded, self.local_dict[self.name], evaluate=False, exact=False)

        if not self.local_dict[self.name] in collected_var.keys() or len(collected_var)>2:
            collected_var = sp.collect(sp.simplify(expanded), self.local_dict[self.name], evaluate=False, exact=False)
            if not self.local_dict[self.name] in collected_var.keys() or len(collected_var)>2:
                Messages._print(self.expression)
                print(collected_var)
                Messages._error('The exponential and event-driven methods are reserved for linear first-order ODEs of the type tau*d'+ self.name+'/dt + '+self.name+' = f(t). Use the explicit method instead.')

        factor_var = collected_var[self.local_dict[self.name]]

        collected_gradient = sp.collect(sp.expand(analysed, grad_var), grad_var, evaluate=False, exact=False)
        if grad_var in collected_gradient.keys():
            factor_gradient = collected_gradient[grad_var]
        else:
            factor_gradient = sp.S(1.0)

        # Real time constant when using the form tau*dV/dt + V = A
        real_tau = factor_gradient / factor_var

        # Normalized equation tau*dV/dt + V = A
        normalized = analysed / factor_var

        # Steady state A
        steadystate = sp.simplify(sp.together(real_tau * grad_var + self.local_dict[self.name] - normalized))

        # Stepsize
        stepsize = sp.simplify(sp.together(self.local_dict['dt']/real_tau))

        return real_tau, stepsize, steadystate

    ###############################################
    ### Conditionals
    ###############################################

    def analyse_condition(self, expression):
        " Analyzes a boolean condition (e.g. for the spike argument)."

        # print("Before:", expression, type(expression))

        expression = expression.strip()

        # Check if there is a == in the condition
        if '==' in expression:
            # Is it the only term, or are there other operations?
            if '&' in expression or '|' in expression:
                expression = re.sub(r'([\w\s.]+)==([\w\s.]+)', r'Equality(\1, \2)', expression)
            else: # only one term
                # Remove brackets
                if expression[0] == "(" and expression[-1] == ")":
                    expression = expression[1:-1]
                terms = expression.split('==')
                expression = 'Equality(' + terms[0] + ', ' + terms[1] + ')'

        # Check if there is a != in the condition
        if '!=' in expression:
            # Is it the only term, or are there other operations?
            if '&' in expression or '|' in expression:
                expression = re.sub(r'([\w\s.]+)!=([\w\s.]+)', r'Not(Equality(\1, \2))', expression)
            else:# only one term
                # Remove brackets
                if expression[0] == "(" and expression[-1] == ")":
                    expression = expression[1:-1]
                terms = expression.split('!=')
                expression = 'Not(Equality(' + terms[0] + ', ' + terms[1] + '))'

        # print("After", expression)

        # Parse the string
        analysed = self.parse_expression(
            expression,
            local_dict = self.local_dict
        )
        self.analysed = analysed
        # print(analysed)

        # Obtain C code
        code = self.c_code(analysed)
        # print(code)

        # Return result
        return code

    ###############################################
    ### Increments
    ###############################################

    def analyse_increment(self, expression):
        " Analyzes an incremental assignment (e.g. a += 0.2)."

        # Get only the right term
        if '+=' in expression:
            name = expression[:expression.find('+=')].strip()
            expression = expression[expression.find('+=')+2:]
            ope = ' += '
        elif '-=' in expression:
            name = expression[:expression.find('-=')].strip()
            expression = expression[expression.find('-=')+2:]
            ope = ' -= '
        elif '*=' in expression:
            name = expression[:expression.find('*=')].strip()
            expression = expression[expression.find('*=')+2:]
            ope = ' *= '
        elif '/=' in expression:
            name = expression[:expression.find('/=')].strip()
            expression = expression[expression.find('/=')+2:]
            ope = ' /= '

        # Parse the string
        analysed = self.parse_expression(
            expression,
            local_dict = self.local_dict
        )
        self.analysed = analysed

        # Obtain C code
        code = self.c_code(self.local_dict[name]) + ope + self.c_code(sp.simplify(analysed, ratio=1.0)) +';'

        # number of floating operations
        self.num_flops = 1 + self._count_ops(self.analysed)

        # Return result
        return code

    ###############################################
    ### Assignments
    ###############################################

    def analyse_assignment(self, expression):
        " Analyzes a simple assignment (e.g. a = 0.2)."

        # Get the left and right term
        name = expression[:expression.find('=')].strip()
        expression = expression[expression.find('=')+1:]

        # Parse the string
        self.analysed = self.parse_expression(
            expression,
            local_dict = self.local_dict
        )

        # Obtain C code
        code = self.c_code(self.local_dict[name]) + ' = ' + self.c_code(self.analysed) +';'

        # number of floating operations
        self.num_flops = self._count_ops(self.analysed)

        # Return result
        return code

    def analyse_return(self, expression):
        " Analyzes a return statement (e.g. w * pre.r)."

        # Parse the string
        analysed = self.parse_expression(
            expression,
            local_dict = self.local_dict
        )
        self.analysed = analysed

        # Obtain C code
        code = self.c_code(analysed) +';'

        # number of floating operations
        self.num_flops = self._count_ops(self.analysed)

        # Return result
        return code

    ###############################################
    ### Dependencies
    ###############################################

    def dependencies(self):
        "Returns all dependencies of the equation"
        deps = []
        for att in self.attributes:
            if self.local_dict[att] in self.analysed.atoms():
                deps.append(att)
        return deps
